# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy import AsyncCollection, AsyncDatabase, AsyncTable, DataAPIClient
from astrapy.authentication import StaticTokenProvider
from astrapy.constants import Environment
from astrapy.exceptions import InvalidEnvironmentException
from astrapy.settings.defaults import DEFAULT_ASTRA_DB_KEYSPACE
from astrapy.utils.api_options import (
    APIOptions,
    DataAPIURLOptions,
    defaultAPIOptions,
)

from ..conftest import (
    TEST_COLLECTION_INSTANCE_NAME,
    TEST_TABLE_INSTANCE_NAME,
    DataAPICredentials,
    DataAPICredentialsInfo,
    DefaultAsyncCollection,
    DefaultAsyncTable,
)

api_ep5643_prod = (
    "https://56439999-89ab-cdef-0123-456789abcdef-region.apps.astra.datastax.com"
)


class TestDatabasesAsync:
    @pytest.mark.describe("test of instantiating Database, async")
    async def test_instantiate_database_async(
        self,
        data_api_credentials_kwargs: DataAPICredentials,
        data_api_credentials_info: DataAPICredentialsInfo,
    ) -> None:
        opts0 = defaultAPIOptions(environment=data_api_credentials_info["environment"])
        opts = opts0.with_override(
            APIOptions(
                token=data_api_credentials_kwargs["token"],
            )
        )

        db1 = AsyncDatabase(
            api_endpoint=data_api_credentials_kwargs["api_endpoint"],
            keyspace=data_api_credentials_kwargs["keyspace"],
            api_options=opts,
        )
        db2 = AsyncDatabase(
            api_endpoint=data_api_credentials_kwargs["api_endpoint"],
            keyspace=data_api_credentials_kwargs["keyspace"],
            api_options=opts,
        )
        assert db1 == db2

    @pytest.mark.describe("test of Database conversions, async")
    async def test_convert_database_async(
        self,
        data_api_credentials_kwargs: DataAPICredentials,
        data_api_credentials_info: DataAPICredentialsInfo,
    ) -> None:
        opts0 = defaultAPIOptions(environment=data_api_credentials_info["environment"])
        opts = opts0.with_override(
            APIOptions(
                token=data_api_credentials_kwargs["token"],
            )
        )

        db1 = AsyncDatabase(
            api_endpoint=data_api_credentials_kwargs["api_endpoint"],
            keyspace=data_api_credentials_kwargs["keyspace"],
            api_options=opts,
        )
        assert db1 == db1._copy()
        assert db1 == db1.with_options()
        assert db1 == db1.to_sync().to_async()

    @pytest.mark.describe("test of Database rich _copy, async")
    async def test_rich_copy_database_async(
        self,
    ) -> None:
        callers0 = [("cn", "cv"), ("dn", "dv")]
        db1 = AsyncDatabase(
            api_endpoint="api_endpoint",
            keyspace="keyspace",
            api_options=defaultAPIOptions(environment="other").with_override(
                APIOptions(
                    token=StaticTokenProvider("token"),
                    callers=callers0,
                    data_api_url_options=DataAPIURLOptions(
                        api_path="api_path",
                        api_version="api_version",
                    ),
                ),
            ),
        )
        assert db1 != db1._copy(token="x")
        assert db1 != db1._copy(keyspace="x")

        db2 = db1._copy(
            token="x",
            keyspace="x",
        )
        assert db2 != db1

        assert db1.with_options(keyspace="x") != db1

        assert db1.with_options(keyspace="x").with_options(keyspace="keyspace") == db1

    @pytest.mark.describe("test of Database rich conversions, async")
    async def test_rich_convert_database_async(
        self,
    ) -> None:
        callers0 = [("cn", "cv"), ("dn", "dv")]
        db1 = AsyncDatabase(
            api_endpoint="api_endpoint",
            keyspace="keyspace",
            api_options=defaultAPIOptions(environment="other").with_override(
                APIOptions(
                    token=StaticTokenProvider("token"),
                    callers=callers0,
                    data_api_url_options=DataAPIURLOptions(
                        api_path="api_path",
                        api_version="api_version",
                    ),
                ),
            ),
        )
        assert db1 != db1.to_sync(token="o").to_async()
        assert db1 != db1.to_sync(keyspace="o").to_async()

        db2s = db1.to_sync(
            token="x",
            keyspace="x",
        )
        assert db2s.to_async() != db1

        db3 = db2s.to_async(
            token="token",
            keyspace="keyspace",
        )
        assert db3 == db1

    @pytest.mark.describe("test get_collection method, async")
    async def test_database_get_collection_async(
        self,
        async_database: AsyncDatabase,
        async_collection_instance: DefaultAsyncCollection,
    ) -> None:
        collection = async_database.get_collection(TEST_COLLECTION_INSTANCE_NAME)
        assert collection == async_collection_instance

        assert getattr(async_database, TEST_COLLECTION_INSTANCE_NAME) == collection
        assert async_database[TEST_COLLECTION_INSTANCE_NAME] == collection

        KEYSPACE_2 = "other_keyspace"
        collection_ks2 = async_database.get_collection(
            TEST_COLLECTION_INSTANCE_NAME, keyspace=KEYSPACE_2
        )
        assert collection_ks2 == AsyncCollection(
            database=async_database,
            name=TEST_COLLECTION_INSTANCE_NAME,
            keyspace=KEYSPACE_2,
            api_options=async_database.api_options,
        )
        assert collection_ks2.database.keyspace == KEYSPACE_2

        kcollection = async_database.get_collection(
            TEST_COLLECTION_INSTANCE_NAME,
            keyspace=KEYSPACE_2,
            embedding_api_key="eak",
            reranking_api_key="eak",
        )
        assert kcollection == AsyncCollection(
            database=async_database,
            name=TEST_COLLECTION_INSTANCE_NAME,
            keyspace=KEYSPACE_2,
            api_options=async_database.api_options.with_override(
                APIOptions(
                    embedding_api_key="eak",
                    reranking_api_key="eak",
                )
            ),
        )

    @pytest.mark.describe("test get_table method, async")
    async def test_database_get_table_async(
        self,
        async_database: AsyncDatabase,
        async_table_instance: DefaultAsyncTable,
        data_api_credentials_kwargs: DataAPICredentials,
    ) -> None:
        table = async_database.get_table(TEST_TABLE_INSTANCE_NAME)
        assert table == async_table_instance

        KEYSPACE_2 = "other_keyspace"
        table_ks2 = async_database.get_table(
            TEST_TABLE_INSTANCE_NAME, keyspace=KEYSPACE_2
        )
        assert table_ks2 == AsyncTable(
            database=async_database,
            name=TEST_TABLE_INSTANCE_NAME,
            keyspace=KEYSPACE_2,
            api_options=async_database.api_options,
        )
        assert table_ks2.database.keyspace == KEYSPACE_2

        ktable = async_database.get_table(
            TEST_TABLE_INSTANCE_NAME,
            keyspace=KEYSPACE_2,
            embedding_api_key="eak",
            reranking_api_key="eak",
        )
        assert ktable == AsyncTable(
            database=async_database,
            name=TEST_TABLE_INSTANCE_NAME,
            keyspace=KEYSPACE_2,
            api_options=async_database.api_options.with_override(
                APIOptions(
                    embedding_api_key="eak",
                    reranking_api_key="eak",
                )
            ),
        )

    @pytest.mark.describe("test database id and region, async")
    async def test_database_id_region_async(self) -> None:
        db1 = AsyncDatabase(
            api_endpoint="https://a1234567-89ab-cdef-0123-456789abcdef-us-central1.apps.astra-dev.datastax.com",
            keyspace="k",
            api_options=defaultAPIOptions(environment="dev"),
        )
        assert db1.id == "a1234567-89ab-cdef-0123-456789abcdef"
        assert db1.region == "us-central1"

        db2 = AsyncDatabase(
            api_endpoint="http://localhost:12345",
            keyspace="k",
            api_options=defaultAPIOptions(environment="dev"),
        )
        with pytest.raises(InvalidEnvironmentException):
            db2.id
        with pytest.raises(InvalidEnvironmentException):
            db2.region

    @pytest.mark.describe("test database default keyspace per environment, async")
    async def test_database_default_keyspace_per_environment_async(self) -> None:
        opts_p = defaultAPIOptions(environment=Environment.PROD)
        opts_o = defaultAPIOptions(environment=Environment.OTHER)
        db_a_m = AsyncDatabase(
            api_endpoint="ep",
            keyspace="M",
            api_options=opts_p,
        )
        assert db_a_m.keyspace == "M"
        db_o_m = AsyncDatabase(
            api_endpoint="ep",
            keyspace="M",
            api_options=opts_o,
        )
        assert db_o_m.keyspace == "M"
        db_a_n = AsyncDatabase(
            api_endpoint="ep",
            keyspace=None,
            api_options=opts_p,
        )
        assert db_a_n.keyspace == DEFAULT_ASTRA_DB_KEYSPACE
        db_o_n = AsyncDatabase(
            api_endpoint="ep",
            keyspace=None,
            api_options=opts_o,
        )
        assert db_o_n.keyspace is None

    @pytest.mark.describe(
        "test database-from-client default keyspace per environment, async"
    )
    async def test_database_from_client_default_keyspace_per_environment_async(
        self,
    ) -> None:
        client_a = DataAPIClient(environment=Environment.PROD)
        db_a_me = client_a.get_async_database(api_ep5643_prod, keyspace="M")
        assert db_a_me.keyspace == "M"
        db_a_ne = client_a.get_async_database(api_ep5643_prod)
        assert db_a_ne.keyspace == DEFAULT_ASTRA_DB_KEYSPACE

        client_o = DataAPIClient(environment=Environment.OTHER)
        db_a_m = client_o.get_async_database("http://a", keyspace="M")
        assert db_a_m.keyspace == "M"
        db_a_n = client_o.get_async_database("http://a")
        assert db_a_n.keyspace is None

    @pytest.mark.describe(
        "test database-from-dataapidbadmin default keyspace per environment, async"
    )
    async def test_database_from_dataapidbadmin_default_keyspace_per_environment_async(
        self,
    ) -> None:
        client = DataAPIClient(environment=Environment.OTHER)
        db_admin = client.get_async_database("http://a").get_database_admin()
        db_m = db_admin.get_async_database(keyspace="M")
        assert db_m.keyspace == "M"
        db_n = db_admin.get_async_database()
        assert db_n.keyspace is None

    @pytest.mark.describe("test of database keyspace property, async")
    def test_database_keyspace_property_async(
        self,
        async_database: AsyncDatabase,
    ) -> None:
        assert isinstance(async_database.keyspace, str)

    @pytest.mark.describe("test of database with custom domains, async")
    async def test_database_custom_domain_async(self) -> None:
        client = DataAPIClient()
        adatabase = client.get_async_database("http://bubbu.com")

        with pytest.raises(InvalidEnvironmentException):
            adatabase.id
        with pytest.raises(InvalidEnvironmentException):
            adatabase.region
        with pytest.raises(InvalidEnvironmentException):
            await adatabase.info()
        with pytest.raises(InvalidEnvironmentException):
            await adatabase.name()
        with pytest.raises(InvalidEnvironmentException):
            adatabase.get_database_admin()



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy.utils.str_enum import StrEnum


class MyTestEnum(StrEnum):
    VALUE = "value"
    VALUE_DASH = "value-dash"


class TestStrEnum:
    def test_strenum_contains(self) -> None:
        assert "value" in MyTestEnum
        assert "value_dash" in MyTestEnum
        assert "value-dash" in MyTestEnum
        assert "VALUE-DASH" in MyTestEnum
        assert "pippo" not in MyTestEnum
        assert {6: 12} not in MyTestEnum

    def test_strenum_coerce(self) -> None:
        MyTestEnum.coerce("value")
        MyTestEnum.coerce("value_dash")
        MyTestEnum.coerce("value-dash")
        MyTestEnum.coerce("VALUE-DASH")
        MyTestEnum.coerce(MyTestEnum.VALUE)
        with pytest.raises(ValueError):
            MyTestEnum.coerce("pippo")



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy import Database
from astrapy.cursors import CursorState
from astrapy.exceptions import (
    CollectionDeleteManyException,
    CollectionInsertManyException,
    CollectionUpdateManyException,
    CursorException,
    DataAPIResponseException,
    TooManyDocumentsToCountException,
)

from ..conftest import IS_ASTRA_DB, DefaultCollection


class TestCollectionExceptionsSync:
    @pytest.mark.describe("test of collection insert_many type-failure modes, sync")
    def test_collection_insert_many_type_failures_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        col = sync_empty_collection
        bad_docs = [{"_id": tid} for tid in ["a", "b", "c", ValueError, "e", "f"]]

        with pytest.raises(ValueError):
            col.insert_many([], ordered=True, concurrency=2)

        with pytest.raises(TypeError):
            col.insert_many(bad_docs, ordered=True, chunk_size=2)

        with pytest.raises(TypeError):
            col.insert_many(bad_docs, ordered=False, chunk_size=2, concurrency=1)

        with pytest.raises(TypeError):
            col.insert_many(bad_docs, ordered=False, chunk_size=2, concurrency=2)

    @pytest.mark.describe("test of collection insert_many insert-failure modes, sync")
    def test_collection_insert_many_insert_failures_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        col = sync_empty_collection
        ok_ids = ["a", "b", "c", "d", "e", "f"]
        ok_docs = [{"_id": tid, "i": i} for i, tid in enumerate(ok_ids)]
        dup_ids = ["a", "b", "b", "d", "a", "b", "e", "f"]
        dup_docs = [{"_id": tid, "i": i} for i, tid in enumerate(dup_ids)]

        im_result1 = col.insert_many(ok_docs, ordered=True, chunk_size=2, concurrency=1)
        assert set(im_result1.inserted_ids) == set(ok_ids)
        assert len(list(col.find({}))) == 6

        col.delete_many({})
        im_result2 = col.insert_many(
            ok_docs, ordered=False, chunk_size=2, concurrency=1
        )
        assert set(im_result2.inserted_ids) == set(ok_ids)
        assert len(list(col.find({}))) == 6

        col.delete_many({})
        im_result3 = col.insert_many(
            ok_docs, ordered=False, chunk_size=2, concurrency=2
        )
        assert set(im_result3.inserted_ids) == set(ok_ids)
        assert len(list(col.find({}))) == 6

        col.delete_many({})
        with pytest.raises(CollectionInsertManyException) as exc:
            col.insert_many(dup_docs, ordered=True, chunk_size=2, concurrency=1)
        assert len(exc.value.exceptions) == 1
        the_exception = exc.value.exceptions[0]
        assert isinstance(the_exception, DataAPIResponseException)
        assert len(the_exception.error_descriptors) == 1
        assert isinstance(the_exception.command, dict)
        assert isinstance(the_exception.raw_response, dict)
        assert exc.value.inserted_ids == ["a", "b"]
        assert {doc["_id"] for doc in col.find()} == {"a", "b"}

        col.delete_many({})
        with pytest.raises(CollectionInsertManyException) as exc:
            col.insert_many(dup_docs, ordered=False, chunk_size=2, concurrency=1)
        assert len(exc.value.exceptions) == 2
        the_exceptions = exc.value.exceptions
        assert all(isinstance(exc, DataAPIResponseException) for exc in the_exceptions)
        assert all(isinstance(exc.command, dict) for exc in the_exceptions)  # type: ignore[attr-defined]
        assert all(isinstance(exc.raw_response, dict) for exc in the_exceptions)  # type: ignore[attr-defined]
        assert the_exceptions[0].command != the_exceptions[1].command  # type: ignore[attr-defined]
        assert the_exceptions[0].raw_response != the_exceptions[1].raw_response  # type: ignore[attr-defined]
        assert (
            len(the_exceptions[0].error_descriptors)  # type: ignore[attr-defined]
            + len(the_exceptions[1].error_descriptors)  # type: ignore[attr-defined]
            == 3
        )
        assert set(exc.value.inserted_ids) == {"a", "b", "d", "e", "f"}
        assert {doc["_id"] for doc in col.find()} == {"a", "b", "d", "e", "f"}

        col.delete_many({})
        with pytest.raises(CollectionInsertManyException) as exc:
            im_result3 = col.insert_many(
                dup_docs, ordered=False, chunk_size=2, concurrency=2
            )
        assert len(exc.value.exceptions) == 2
        the_exceptions = exc.value.exceptions
        assert all(isinstance(exc, DataAPIResponseException) for exc in the_exceptions)
        assert all(isinstance(exc.command, dict) for exc in the_exceptions)  # type: ignore[attr-defined]
        assert all(isinstance(exc.raw_response, dict) for exc in the_exceptions)  # type: ignore[attr-defined]
        assert the_exceptions[0].command != the_exceptions[1].command  # type: ignore[attr-defined]
        assert the_exceptions[0].raw_response != the_exceptions[1].raw_response  # type: ignore[attr-defined]
        assert (
            len(the_exceptions[0].error_descriptors)  # type: ignore[attr-defined]
            + len(the_exceptions[1].error_descriptors)  # type: ignore[attr-defined]
            == 3
        )
        assert set(exc.value.inserted_ids) == {"a", "b", "d", "e", "f"}
        assert {doc["_id"] for doc in col.find()} == {"a", "b", "d", "e", "f"}

    @pytest.mark.describe("test of collection insert_one failure modes, sync")
    def test_collection_insert_one_failures_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        col = sync_empty_collection
        with pytest.raises(TypeError):
            col.insert_one({"a": ValueError})

        col.insert_one({"_id": "a"})
        with pytest.raises(DataAPIResponseException) as exc:
            col.insert_one({"_id": "a"})
        assert len(exc.value.error_descriptors) == 1
        assert isinstance(exc.value.command, dict)

    @pytest.mark.describe("test of collection options failure modes, sync")
    def test_collection_options_failures_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        col = sync_empty_collection._copy()
        col._name += "_hacked"
        with pytest.raises(RuntimeError, match="not found"):
            col.options()

    @pytest.mark.describe("test of collection count_documents failure modes, sync")
    def test_collection_count_documents_failures_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        col = sync_empty_collection._copy()
        col._name += "_hacked"
        col._api_commander.full_path += "_hacked"
        with pytest.raises(DataAPIResponseException):
            col.count_documents({}, upper_bound=1)
        sync_empty_collection.insert_one({"a": 1})
        sync_empty_collection.insert_one({"b": 2})
        assert sync_empty_collection.count_documents({}, upper_bound=3) == 2
        with pytest.raises(TooManyDocumentsToCountException) as exc:
            sync_empty_collection.count_documents({}, upper_bound=1)
        assert not exc.value.server_max_count_exceeded

    @pytest.mark.describe("test of collection bulk-method DML failure modes, sync")
    def test_collection_bulkmethods_dml_failures_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        col = sync_empty_collection._copy()
        col._name += "_hacked"
        col._api_commander.full_path += "_hacked"
        with pytest.raises(CollectionDeleteManyException):
            col.delete_many({})
        with pytest.raises(CollectionUpdateManyException):
            col.update_many({}, update={"$set": {"a": 1}})

    @pytest.mark.describe("test of collection one-doc DML failure modes, sync")
    def test_collection_monodoc_dml_failures_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        col = sync_empty_collection._copy()
        col._name += "_hacked"
        col._api_commander.full_path += "_hacked"
        with pytest.raises(DataAPIResponseException):
            col.delete_one({"a": 1})
        with pytest.raises(DataAPIResponseException):
            col.find_one_and_replace({"a": 1}, {"a": -1})
        with pytest.raises(DataAPIResponseException):
            col.find_one_and_update({"a": 1}, {"$set": {"a": -1}})
        with pytest.raises(DataAPIResponseException):
            col.find_one_and_delete({"a": 1})
        with pytest.raises(DataAPIResponseException):
            col.replace_one({"a": 1}, {"a": -1})
        with pytest.raises(DataAPIResponseException):
            col.update_one({"a": 1}, {"$set": {"a": -1}})

    @pytest.mark.describe("test of database one-request method failures, sync")
    def test_database_method_failures_sync(
        self,
        sync_database: Database,
    ) -> None:
        f_database = sync_database._copy(keyspace="nonexisting")
        if IS_ASTRA_DB:
            with pytest.raises(DataAPIResponseException):
                f_database.drop_collection("nonexisting")
        with pytest.raises(DataAPIResponseException):
            sync_database.command(body={"myCommand": {"k": "v"}})
        with pytest.raises(DataAPIResponseException):
            sync_database.command(body={"myCommand": {"k": "v"}}, keyspace="ns")
        with pytest.raises(DataAPIResponseException):
            sync_database.command(
                body={"myCommand": {"k": "v"}},
                keyspace="ns",
                collection_or_table_name="coll",
            )
        with pytest.raises(DataAPIResponseException):
            list(sync_database.list_collections(keyspace="nonexisting"))
        with pytest.raises(DataAPIResponseException):
            sync_database.list_collection_names(keyspace="nonexisting")

    @pytest.mark.describe("test of hard exceptions in cursors, sync")
    def test_cursor_hard_exceptions_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        with pytest.raises(TypeError):
            sync_empty_collection.distinct(
                "a",
                filter={"f": ValueError("nonserializable")},
            )

    @pytest.mark.describe("test of custom exceptions in cursors, sync")
    def test_cursor_custom_exceptions_sync(
        self,
        sync_empty_collection: DefaultCollection,
    ) -> None:
        sync_empty_collection.insert_many([{"a": 1}] * 4)
        cur1 = sync_empty_collection.find({})
        cur1.limit(10)

        cur1.__next__()
        with pytest.raises(CursorException) as exc:
            cur1.limit(1)
        assert exc.value.cursor_state == CursorState.STARTED.value

        list(cur1)
        with pytest.raises(CursorException) as exc:
            cur1.limit(1)
        assert exc.value.cursor_state == CursorState.CLOSED.value

    @pytest.mark.describe("test of standard exceptions in cursors, sync")
    def test_cursor_standard_exceptions_sync(self, sync_database: Database) -> None:
        wcol = sync_database.get_collection("nonexisting")
        cur1 = wcol.find({})
        cur2 = wcol.find({})

        with pytest.raises(DataAPIResponseException):
            for item in cur1:
                pass

        with pytest.raises(DataAPIResponseException):
            cur2.__next__()

        with pytest.raises(DataAPIResponseException):
            wcol.distinct("f")

        with pytest.raises(DataAPIResponseException):
            wcol.find_one({})

    @pytest.mark.describe("test of exceptions in list_collections, sync")
    def test_list_collections_hard_exceptions_sync(
        self,
        sync_database: Database,
    ) -> None:
        with pytest.raises(DataAPIResponseException):
            sync_database.list_collections(keyspace="nonexisting")



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Iterable, TypedDict

import pytest

from astrapy import AsyncDatabase, AsyncTable, Database, Table
from astrapy.info import AlterTableAddColumns, AlterTableDropColumns

from ..conftest import DefaultAsyncTable, DefaultTable


class MyTestDoc(TypedDict):
    p_bigint: int
    p_ascii: str


class AlteredMyTestDoc(TypedDict):
    p_bigint: int
    p_ascii: str
    added: int


class MyTestMiniDoc(TypedDict):
    p_bigint: int


FIND_PROJECTION = {"p_bigint": True}
TYPING_TABLE_NAME = "test_typing_table"
TYPING_TABLE_DEFINITION = {
    "columns": {
        "p_ascii": "text",
        "p_bigint": "bigint",
        "p_float": "float",
    },
    "primaryKey": {
        "partitionBy": ["p_ascii", "p_bigint"],
        "partitionSort": {},
    },
}
ROW = {"p_ascii": "abc", "p_bigint": 10000, "p_float": 0.123}
TYPED_ROW: MyTestDoc = {"p_ascii": "abc", "p_bigint": 10000}
FIND_FILTER = {"p_ascii": "abc", "p_bigint": 10000}

ALTER_ADD_COLUMN = AlterTableAddColumns.coerce({"columns": {"added": "int"}})
ALTER_DROP_COLUMN = AlterTableDropColumns.coerce({"columns": ["added"]})


@pytest.fixture(scope="module")
def sync_typing_test_table(sync_database: Database) -> Iterable[DefaultTable]:
    table = sync_database.create_table(
        TYPING_TABLE_NAME,
        definition=TYPING_TABLE_DEFINITION,
        if_not_exists=True,
    )
    yield table

    table.drop()


@pytest.fixture
def async_typing_test_table(
    sync_typing_test_table: DefaultTable,
) -> DefaultAsyncTable:
    return sync_typing_test_table.to_async()


class TestTableTyping:
    @pytest.mark.describe("test of typing create_table and alter_table, sync")
    def test_create_alter_table_typing_sync(
        self,
        sync_database: Database,
        sync_typing_test_table: DefaultTable,
    ) -> None:
        """Test of creating typed tables with generics (and not), sync."""

        # Untyped baseline
        c_tb_untyped = sync_database.create_table(
            TYPING_TABLE_NAME,
            definition=TYPING_TABLE_DEFINITION,
            if_not_exists=True,
        )
        c_tb_untyped.insert_one(ROW)
        cu_doc = c_tb_untyped.find_one(FIND_FILTER)
        assert cu_doc is not None
        cu_a: str
        cu_b: int
        cu_a = cu_doc["p_ascii"]  # noqa: F841
        cu_b = cu_doc["p_bigint"]  # noqa: F841
        assert set(cu_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        cu_x: int
        cu_y: float
        cu_x = cu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            cu_y = cu_doc["c"]  # noqa: F841

        # untyped, cursors type inference on find
        c_tb_untyped_cursor = c_tb_untyped.find({}, projection=FIND_PROJECTION)
        cucur_doc = c_tb_untyped_cursor.__next__()
        assert cucur_doc is not None
        cucur_b: int
        cucur_b = cucur_doc["p_bigint"]  # noqa: F841
        assert set(cucur_doc.keys()) == {"p_bigint"}
        # untyped, these are all ok:
        cucur_x: str
        cucur_y: float
        cucur_x = cucur_doc["p_bigint"]  # noqa: F841
        with pytest.raises(KeyError):
            cucur_y = cucur_doc["c"]  # noqa: F841

        # untyped, check using alter table's return value
        altered_c_tb_untyped = c_tb_untyped.alter(ALTER_ADD_COLUMN)
        altered_cu_doc = altered_c_tb_untyped.find_one(FIND_FILTER)
        assert altered_cu_doc is not None
        alt_cu_a: str
        alt_cu_b: int
        alt_cu_a = altered_cu_doc["added"]  # noqa: F841
        alt_cu_b = altered_cu_doc["added"]  # noqa: F841
        altered_c_tb_untyped.alter(ALTER_DROP_COLUMN)

        # Typed
        c_tb_typed: Table[MyTestDoc] = sync_database.create_table(
            TYPING_TABLE_NAME,
            definition=TYPING_TABLE_DEFINITION,
            if_not_exists=True,
            row_type=MyTestDoc,
        )
        c_tb_typed.insert_one(TYPED_ROW)
        ct_doc = c_tb_typed.find_one(FIND_FILTER)
        assert ct_doc is not None
        ct_a: str
        ct_b: int
        ct_a = ct_doc["p_ascii"]  # noqa: F841
        ct_b = ct_doc["p_bigint"]  # noqa: F841
        assert set(ct_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ct_x: int
        ct_y: float
        ct_x = ct_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ct_y = ct_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, cursors type inference on find
        c_tb_typed_cursor = c_tb_typed.find(
            {}, projection=FIND_PROJECTION, row_type=MyTestMiniDoc
        )
        ctcur_doc = c_tb_typed_cursor.__next__()
        assert ctcur_doc is not None
        ctcur_b: int
        ctcur_b = ctcur_doc["p_bigint"]  # noqa: F841
        assert set(ctcur_doc.keys()) == {"p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ctcur_x: str
        ctcur_y: float
        ctcur_x = ctcur_doc["p_bigint"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ctcur_y = ctcur_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, check using alter table's return value
        altered_c_tb_typed = c_tb_typed.alter(
            ALTER_ADD_COLUMN, row_type=AlteredMyTestDoc
        )
        altered_ct_doc = altered_c_tb_typed.find_one(FIND_FILTER)
        assert altered_ct_doc is not None
        alt_ct_a: str
        alt_ct_b: int
        alt_ct_a = altered_ct_doc["added"]  # type: ignore[assignment] # noqa: F841
        alt_ct_b = altered_ct_doc["added"]  # noqa: F841
        altered_c_tb_typed.alter(ALTER_DROP_COLUMN)

    @pytest.mark.describe("test of typing get_table, sync")
    def test_get_table_typing_sync(
        self,
        sync_database: Database,
        sync_typing_test_table: DefaultTable,
    ) -> None:
        """Test of getting typed tables with generics (and not), sync."""

        # Untyped baseline
        g_tb_untyped = sync_database.get_table(TYPING_TABLE_NAME)
        g_tb_untyped.insert_one(ROW)
        gu_doc = g_tb_untyped.find_one(FIND_FILTER)
        assert gu_doc is not None
        gu_a: str
        gu_b: int
        gu_a = gu_doc["p_ascii"]  # noqa: F841
        gu_b = gu_doc["p_bigint"]  # noqa: F841
        assert set(gu_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        gu_x: int
        gu_y: float
        gu_x = gu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            gu_y = gu_doc["c"]  # noqa: F841

        # Typed
        g_tb_typed: Table[MyTestDoc] = sync_database.get_table(
            TYPING_TABLE_NAME, row_type=MyTestDoc
        )
        g_tb_typed.insert_one(TYPED_ROW)
        gt_doc = g_tb_typed.find_one(FIND_FILTER)
        assert gt_doc is not None
        gt_a: str
        gt_b: int
        gt_a = gt_doc["p_ascii"]  # noqa: F841
        gt_b = gt_doc["p_bigint"]  # noqa: F841
        assert set(gt_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        gt_x: int
        gt_y: float
        gt_x = gt_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_y = gt_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

    @pytest.mark.describe("test of typing create_table and alter_table, async")
    async def test_create_alter_table_typing_async(
        self,
        async_database: AsyncDatabase,
        async_typing_test_table: DefaultAsyncTable,
    ) -> None:
        """Test of creating typed tables with generics (and not), async."""

        # Untyped baseline
        ac_tb_untyped = await async_database.create_table(
            TYPING_TABLE_NAME,
            definition=TYPING_TABLE_DEFINITION,
            if_not_exists=True,
        )
        await ac_tb_untyped.insert_one(ROW)
        cu_doc = await ac_tb_untyped.find_one(FIND_FILTER)
        assert cu_doc is not None
        cu_a: str
        cu_b: int
        cu_a = cu_doc["p_ascii"]  # noqa: F841
        cu_b = cu_doc["p_bigint"]  # noqa: F841
        assert set(cu_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        cu_x: int
        cu_y: float
        cu_x = cu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            cu_y = cu_doc["c"]  # noqa: F841

        # untyped, cursors type inference on find
        c_tb_untyped_acursor = ac_tb_untyped.find({}, projection=FIND_PROJECTION)
        cucur_doc = await c_tb_untyped_acursor.__anext__()
        assert cucur_doc is not None
        cucur_b: int
        cucur_b = cucur_doc["p_bigint"]  # noqa: F841
        assert set(cucur_doc.keys()) == {"p_bigint"}
        # untyped, these are all ok:
        cucur_x: str
        cucur_y: float
        cucur_x = cucur_doc["p_bigint"]  # noqa: F841
        with pytest.raises(KeyError):
            cucur_y = cucur_doc["c"]  # noqa: F841

        # untyped, check using alter table's return value
        altered_ac_tb_untyped = await ac_tb_untyped.alter(ALTER_ADD_COLUMN)
        altered_cu_doc = await altered_ac_tb_untyped.find_one(FIND_FILTER)
        assert altered_cu_doc is not None
        alt_cu_a: str
        alt_cu_b: int
        alt_cu_a = altered_cu_doc["added"]  # noqa: F841
        alt_cu_b = altered_cu_doc["added"]  # noqa: F841
        await altered_ac_tb_untyped.alter(ALTER_DROP_COLUMN)

        # Typed
        ac_tb_typed: AsyncTable[MyTestDoc] = await async_database.create_table(
            TYPING_TABLE_NAME,
            definition=TYPING_TABLE_DEFINITION,
            if_not_exists=True,
            row_type=MyTestDoc,
        )
        await ac_tb_typed.insert_one(TYPED_ROW)
        ct_doc = await ac_tb_typed.find_one(FIND_FILTER)
        assert ct_doc is not None
        ct_a: str
        ct_b: int
        ct_a = ct_doc["p_ascii"]  # noqa: F841
        ct_b = ct_doc["p_bigint"]  # noqa: F841
        assert set(ct_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ct_x: int
        ct_y: float
        ct_x = ct_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ct_y = ct_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, cursors type inference on find
        c_tb_typed_acursor = ac_tb_typed.find(
            {}, projection=FIND_PROJECTION, row_type=MyTestMiniDoc
        )
        ctcur_doc = await c_tb_typed_acursor.__anext__()
        assert ctcur_doc is not None
        ctcur_b: int
        ctcur_b = ctcur_doc["p_bigint"]  # noqa: F841
        assert set(ctcur_doc.keys()) == {"p_bigint"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        ctcur_x: str
        ctcur_y: float
        ctcur_x = ctcur_doc["p_bigint"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            ctcur_y = ctcur_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841

        # typed, check using alter table's return value
        altered_ac_tb_typed = await ac_tb_typed.alter(
            ALTER_ADD_COLUMN, row_type=AlteredMyTestDoc
        )
        altered_ct_doc = await altered_ac_tb_typed.find_one(FIND_FILTER)
        assert altered_ct_doc is not None
        alt_ct_a: str
        alt_ct_b: int
        alt_ct_a = altered_ct_doc["added"]  # type: ignore[assignment] # noqa: F841
        alt_ct_b = altered_ct_doc["added"]  # noqa: F841
        await altered_ac_tb_typed.alter(ALTER_DROP_COLUMN)

    @pytest.mark.describe("test of typing get_table, async")
    async def test_get_table_typing_async(
        self,
        async_database: AsyncDatabase,
        async_typing_test_table: DefaultAsyncTable,
    ) -> None:
        """Test of getting typed tables with generics (and not), async."""

        # Untyped baseline
        ag_tb_untyped = async_database.get_table(TYPING_TABLE_NAME)
        await ag_tb_untyped.insert_one(ROW)
        gu_doc = await ag_tb_untyped.find_one(FIND_FILTER)
        assert gu_doc is not None
        gu_a: str
        gu_b: int
        gu_a = gu_doc["p_ascii"]  # noqa: F841
        gu_b = gu_doc["p_bigint"]  # noqa: F841
        assert set(gu_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # untyped, these are all ok:
        gu_x: int
        gu_y: float
        gu_x = gu_doc["p_ascii"]  # noqa: F841
        with pytest.raises(KeyError):
            gu_y = gu_doc["c"]  # noqa: F841

        # Typed
        ag_tb_typed: AsyncTable[MyTestDoc] = async_database.get_table(
            TYPING_TABLE_NAME, row_type=MyTestDoc
        )
        await ag_tb_typed.insert_one(TYPED_ROW)
        gt_doc = await ag_tb_typed.find_one(FIND_FILTER)
        assert gt_doc is not None
        gt_a: str
        gt_b: int
        gt_a = gt_doc["p_ascii"]  # noqa: F841
        gt_b = gt_doc["p_bigint"]  # noqa: F841
        assert set(gt_doc.keys()) == {"p_ascii", "p_bigint", "p_float"}
        # these two SHOULD NOT typecheck (i.e. require the ignore directive)
        gt_x: int
        gt_y: float
        gt_x = gt_doc["p_ascii"]  # type: ignore[assignment]  # noqa: F841
        with pytest.raises(KeyError):
            gt_y = gt_doc["c"]  # type: ignore[typeddict-item]  # noqa: F841



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json

import pytest

from astrapy.data.table import map2tuple_checker_insert_one
from astrapy.data.utils.table_converters import preprocess_table_payload
from astrapy.data_types import (
    DataAPIDictUDT,
    DataAPIMap,
    DataAPISet,
)
from astrapy.utils.api_options import (
    FullSerdesOptions,
    SerdesOptions,
    defaultSerdesOptions,
)

from ..table_udt_assets import (
    THE_BYTES,
    THE_SERIALIZED_BYTES,
    THE_SERIALIZED_TIMESTAMP,
    THE_TIMESTAMP,
    UnitExtendedPlayer,
    _unit_extended_player_serializer,
)

OPTIONS_DAM = defaultSerdesOptions.with_override(
    SerdesOptions(encode_maps_as_lists_in_tables="DATAAPIMAPS"),
)
OPTIONS_NEV = defaultSerdesOptions.with_override(
    SerdesOptions(encode_maps_as_lists_in_tables="NEVER"),
)
OPTIONS_ALW = defaultSerdesOptions.with_override(
    SerdesOptions(encode_maps_as_lists_in_tables="ALWAYS"),
)

FULL_WRAPPABLE_DICT = {
    "name": "Jamie",
    "age": "45",
    "blb": THE_BYTES,
    "ts": THE_TIMESTAMP,
}
FULL_WRAPPED_DICT = DataAPIDictUDT(FULL_WRAPPABLE_DICT)
FULL_EXTENDEDPLAYER = UnitExtendedPlayer(**FULL_WRAPPABLE_DICT)  # type: ignore[arg-type]
FULL_SERIALIZED_DICT = {
    "name": "Jamie",
    "age": "45",
    "blb": THE_SERIALIZED_BYTES,
    "ts": THE_SERIALIZED_TIMESTAMP,
}
FULL_SERIALIZED_TUPLIFIED_DICT = [[k, v] for k, v in FULL_SERIALIZED_DICT.items()]
FULL_DICTWRAPPER_ROW = {
    "d": FULL_WRAPPABLE_DICT,
    "wr": FULL_WRAPPED_DICT,
    "dam": DataAPIMap([("k", "v")]),
    "dami": DataAPIMap([(1, "v1")]),
    "das_wr": DataAPISet([FULL_WRAPPED_DICT]),
    "dam_wr": DataAPIMap([("k1", FULL_WRAPPED_DICT)]),
    "l_wr": [FULL_WRAPPED_DICT],
    "m_wr": {"k2": FULL_WRAPPED_DICT},
}
FULL_CUSTOMCLASS_ROW = {
    "d": FULL_WRAPPABLE_DICT,
    "wr": FULL_EXTENDEDPLAYER,
    "dam": DataAPIMap([("k", "v")]),
    "dami": DataAPIMap([(1, "v1")]),
    "das_wr": DataAPISet([FULL_EXTENDEDPLAYER]),
    "dam_wr": DataAPIMap([("k1", FULL_EXTENDEDPLAYER)]),
    "l_wr": [FULL_EXTENDEDPLAYER],
    "m_wr": {"k2": FULL_EXTENDEDPLAYER},
}
FULL_SERIALIZED_ROW_DAM = {
    "d": FULL_SERIALIZED_DICT,
    "wr": FULL_SERIALIZED_DICT,
    "dam": [["k", "v"]],
    "dami": [[1, "v1"]],
    "das_wr": [FULL_SERIALIZED_DICT],
    "dam_wr": [["k1", FULL_SERIALIZED_DICT]],
    "l_wr": [FULL_SERIALIZED_DICT],
    "m_wr": {"k2": FULL_SERIALIZED_DICT},
}
FULL_SERIALIZED_ROW_NEV = {
    "d": FULL_SERIALIZED_DICT,
    "wr": FULL_SERIALIZED_DICT,
    "dam": {"k": "v"},
    "dami": {1: "v1"},
    "das_wr": [FULL_SERIALIZED_DICT],
    "dam_wr": {"k1": FULL_SERIALIZED_DICT},
    "l_wr": [FULL_SERIALIZED_DICT],
    "m_wr": {"k2": FULL_SERIALIZED_DICT},
}
FULL_SERIALIZED_ROW_ALW = {
    "d": FULL_SERIALIZED_TUPLIFIED_DICT,
    "wr": FULL_SERIALIZED_DICT,
    "dam": [["k", "v"]],
    "dami": [[1, "v1"]],
    "das_wr": [FULL_SERIALIZED_DICT],
    "dam_wr": [["k1", FULL_SERIALIZED_DICT]],
    "l_wr": [FULL_SERIALIZED_DICT],
    "m_wr": [["k2", FULL_SERIALIZED_DICT]],
}


class TestTPreprocessorsUserDefinedTypes:
    @pytest.mark.describe("Sanity check on map-as-lists default serdes options")
    def test_udt_serdesoptions_sanitycheck(self) -> None:
        assert OPTIONS_DAM == defaultSerdesOptions

    @pytest.mark.describe("test of udt conversion in preprocessing for DataAPIDictUDT")
    def test_udt_preprocessing_dataapidictudt(self) -> None:
        preprocessed_dam = preprocess_table_payload(
            {"insertOne": {"document": FULL_DICTWRAPPER_ROW}},
            OPTIONS_DAM,
            map2tuple_checker_insert_one,
        )["insertOne"]["document"]  # type: ignore[index]
        preprocessed_nev = preprocess_table_payload(
            {"insertOne": {"document": FULL_DICTWRAPPER_ROW}},
            OPTIONS_NEV,
            map2tuple_checker_insert_one,
        )["insertOne"]["document"]  # type: ignore[index]
        preprocessed_alw = preprocess_table_payload(
            {"insertOne": {"document": FULL_DICTWRAPPER_ROW}},
            OPTIONS_ALW,
            map2tuple_checker_insert_one,
        )["insertOne"]["document"]  # type: ignore[index]

        assert preprocessed_dam == FULL_SERIALIZED_ROW_DAM
        assert preprocessed_nev == FULL_SERIALIZED_ROW_NEV
        assert preprocessed_alw == FULL_SERIALIZED_ROW_ALW

        _ = json.dumps(preprocessed_dam)
        _ = json.dumps(preprocessed_nev)
        _ = json.dumps(preprocessed_alw)

    @pytest.mark.describe(
        "test of udt conversion in preprocessing for unregistered class",
    )
    def test_udt_preprocessing_unregisteredclass(self) -> None:
        preprocessed_dam = preprocess_table_payload(
            {"insertOne": {"document": FULL_CUSTOMCLASS_ROW}},
            OPTIONS_DAM,
            map2tuple_checker_insert_one,
        )["insertOne"]["document"]  # type: ignore[index]
        preprocessed_nev = preprocess_table_payload(
            {"insertOne": {"document": FULL_CUSTOMCLASS_ROW}},
            OPTIONS_NEV,
            map2tuple_checker_insert_one,
        )["insertOne"]["document"]  # type: ignore[index]
        preprocessed_alw = preprocess_table_payload(
            {"insertOne": {"document": FULL_CUSTOMCLASS_ROW}},
            OPTIONS_ALW,
            map2tuple_checker_insert_one,
        )["insertOne"]["document"]  # type: ignore[index]

        with pytest.raises(TypeError, match="JSON serializable"):
            json.dumps(preprocessed_dam)
        with pytest.raises(TypeError, match="JSON serializable"):
            json.dumps(preprocessed_nev)
        with pytest.raises(TypeError, match="JSON serializable"):
            json.dumps(preprocessed_alw)

    @pytest.mark.describe(
        "test of udt conversion in preprocessing for a registered class",
    )
    def test_udt_preprocessing_registeredclass(self) -> None:
        def _register_options(api_o: FullSerdesOptions) -> FullSerdesOptions:
            return api_o.with_override(
                SerdesOptions(
                    serializer_by_class={
                        UnitExtendedPlayer: _unit_extended_player_serializer,
                    },
                )
            )

        preprocessed_dam = preprocess_table_payload(
            {"insertOne": {"document": FULL_CUSTOMCLASS_ROW}},
            _register_options(OPTIONS_DAM),
            map2tuple_checker_insert_one,
        )["insertOne"]["document"]  # type: ignore[index]
        preprocessed_nev = preprocess_table_payload(
            {"insertOne": {"document": FULL_CUSTOMCLASS_ROW}},
            _register_options(OPTIONS_NEV),
            map2tuple_checker_insert_one,
        )["insertOne"]["document"]  # type: ignore[index]
        preprocessed_alw = preprocess_table_payload(
            {"insertOne": {"document": FULL_CUSTOMCLASS_ROW}},
            _register_options(OPTIONS_ALW),
            map2tuple_checker_insert_one,
        )["insertOne"]["document"]  # type: ignore[index]

        assert preprocessed_dam == FULL_SERIALIZED_ROW_DAM
        assert preprocessed_nev == FULL_SERIALIZED_ROW_NEV
        assert preprocessed_alw == FULL_SERIALIZED_ROW_ALW

        _ = json.dumps(preprocessed_dam)
        _ = json.dumps(preprocessed_nev)
        _ = json.dumps(preprocessed_alw)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from ..collection_decimal_support_assets import (
    BASELINE_OBJ,
    S_OPTS_NO_DECS,
    S_OPTS_OK_DECS,
    WDECS_OBJ,
    is_decimal_super,
)

__all__ = [
    "BASELINE_OBJ",
    "S_OPTS_NO_DECS",
    "S_OPTS_OK_DECS",
    "WDECS_OBJ",
    "is_decimal_super",
]



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import logging

import pytest
from pytest_httpserver import HTTPServer

from astrapy.exceptions import (
    DataAPIHttpException,
    DataAPIResponseException,
    DevOpsAPIHttpException,
    DevOpsAPIResponseException,
    UnexpectedDataAPIResponseException,
    UnexpectedDevOpsAPIResponseException,
)
from astrapy.utils.api_commander import APICommander
from astrapy.utils.request_tools import HttpMethod


class TestAPICommander:
    @pytest.mark.describe("test of APICommander conversion methods")
    def test_apicommander_conversions(self) -> None:
        cmd1 = APICommander(
            api_endpoint="api_endpoint1",
            path="path1",
            headers={"h": "headers1"},
            callers=[("c", "v")],
            redacted_header_names=["redacted_header_names1"],
            dev_ops_api=True,
        )
        cmd2 = APICommander(
            api_endpoint="api_endpoint1",
            path="path1",
            headers={"h": "headers1"},
            callers=[("c", "v")],
            redacted_header_names=["redacted_header_names1"],
            dev_ops_api=True,
        )
        assert cmd1 == cmd2

        assert cmd1 != cmd1._copy(api_endpoint="x")
        assert cmd1 != cmd1._copy(path="x")
        assert cmd1 != cmd1._copy(headers={})
        assert cmd1 != cmd1._copy(callers=[])
        assert cmd1 != cmd1._copy(redacted_header_names=[])
        assert cmd1 != cmd1._copy(dev_ops_api=False)

        assert cmd1 == cmd1._copy(api_endpoint="x")._copy(api_endpoint="api_endpoint1")
        assert cmd1 == cmd1._copy(path="x")._copy(path="path1")
        assert cmd1 == cmd1._copy(headers={})._copy(headers={"h": "headers1"})
        assert cmd1 == cmd1._copy(callers=[])._copy(callers=[("c", "v")])
        assert cmd1 == cmd1._copy(redacted_header_names=[])._copy(
            redacted_header_names=["redacted_header_names1"]
        )
        assert cmd1 == cmd1._copy(dev_ops_api=False)._copy(dev_ops_api=True)

    @pytest.mark.describe("test of APICommander request, sync")
    def test_apicommander_request_sync(self, httpserver: HTTPServer) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        extra_path = "extra/path"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            headers={"h": "v"},
            callers=[("cn0", "cv0"), ("cn1", "cv1")],
        )

        def hv_matcher(hk: str, hv: str | None, ev: str) -> bool:
            if hk == "v":
                return hv == ev
            elif hk.lower() == "user-agent":
                return hv is not None and hv.startswith(ev)
            else:
                return True

        httpserver.expect_oneshot_request(
            base_path,
            method=HttpMethod.PUT,
            headers={
                "h": "v",
                "User-Agent": "cn0/cv0 cn1/cv1",
            },
            header_value_matcher=hv_matcher,
            data="{}",
        ).respond_with_json({"r": 1})
        resp_b = cmd.request(
            http_method=HttpMethod.PUT,
            payload={},
        )
        assert resp_b == {"r": 1}

        httpserver.expect_oneshot_request(
            "/".join([base_path, extra_path]),
            method=HttpMethod.DELETE,
            data="{}",
        ).respond_with_json({"r": 2})
        resp_e = cmd.request(
            http_method=HttpMethod.DELETE,
            payload={},
            additional_path=extra_path,
        )
        assert resp_e == {"r": 2}

    @pytest.mark.describe("test of APICommander request, async")
    async def test_apicommander_request_async(self, httpserver: HTTPServer) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        extra_path = "extra/path"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            headers={"h": "v"},
            callers=[("cn0", "cv0"), ("cn1", "cv1")],
        )

        def hv_matcher(hk: str, hv: str | None, ev: str) -> bool:
            if hk == "v":
                return hv == ev
            elif hk.lower() == "user-agent":
                return hv is not None and hv.startswith(ev)
            else:
                return True

        httpserver.expect_oneshot_request(
            base_path,
            method=HttpMethod.PUT,
            data="{}",
            headers={
                "h": "v",
                "User-Agent": "cn0/cv0 cn1/cv1",
            },
            header_value_matcher=hv_matcher,
        ).respond_with_json({"r": 1})
        resp_b = await cmd.async_request(
            http_method=HttpMethod.PUT,
            payload={},
        )
        assert resp_b == {"r": 1}

        httpserver.expect_oneshot_request(
            "/".join([base_path, extra_path]),
            method=HttpMethod.DELETE,
            data="{}",
        ).respond_with_json({"r": 2})
        resp_e = await cmd.async_request(
            http_method=HttpMethod.DELETE,
            payload={},
            additional_path=extra_path,
        )
        assert resp_e == {"r": 2}

    @pytest.mark.describe("test of APICommander exceptions, sync")
    def test_apicommander_exceptions_sync(self, httpserver: HTTPServer) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            dev_ops_api=False,
        )

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data("{unparseable")
        with pytest.raises(UnexpectedDataAPIResponseException):
            cmd.request()

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            )
        )
        with pytest.raises(DataAPIResponseException):
            cmd.request()

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            )
        )
        cmd.request(raise_api_errors=False)

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            ),
            status=500,
        )
        with pytest.raises(DataAPIHttpException):
            cmd.request()

    @pytest.mark.describe("test of APICommander exceptions, async")
    async def test_apicommander_exceptions_async(self, httpserver: HTTPServer) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            dev_ops_api=False,
        )

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data("{unparseable")
        with pytest.raises(UnexpectedDataAPIResponseException):
            await cmd.async_request()

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            )
        )
        with pytest.raises(DataAPIResponseException):
            await cmd.async_request()

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            )
        )
        await cmd.async_request(raise_api_errors=False)

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            ),
            status=500,
        )
        with pytest.raises(DataAPIHttpException):
            await cmd.async_request()

    @pytest.mark.describe("test of APICommander DevOps exceptions, sync")
    def test_apicommander_devops_exceptions_sync(self, httpserver: HTTPServer) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            dev_ops_api=True,
        )

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data("{unparseable")
        with pytest.raises(UnexpectedDevOpsAPIResponseException):
            cmd.request()

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            )
        )
        with pytest.raises(DevOpsAPIResponseException):
            cmd.request()

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            )
        )
        cmd.request(raise_api_errors=False)

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            ),
            status=500,
        )
        with pytest.raises(DevOpsAPIHttpException):
            cmd.request()

    @pytest.mark.describe("test of APICommander DevOps exceptions, async")
    async def test_apicommander_devops_exceptions_async(
        self, httpserver: HTTPServer
    ) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            dev_ops_api=True,
        )

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data("{unparseable")
        with pytest.raises(UnexpectedDevOpsAPIResponseException):
            await cmd.async_request()

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            )
        )
        with pytest.raises(DevOpsAPIResponseException):
            await cmd.async_request()

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            )
        )
        await cmd.async_request(raise_api_errors=False)

        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_data(
            json.dumps(
                {
                    "errors": [
                        {"title": "Error", "errorCode": "E_C"},
                    ]
                }
            ),
            status=500,
        )
        with pytest.raises(DevOpsAPIHttpException):
            await cmd.async_request()

    @pytest.mark.describe("test of APICommander server warnings, sync")
    def test_apicommander_server_warnings_sync(
        self,
        caplog: pytest.LogCaptureFixture,
        httpserver: HTTPServer,
    ) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            dev_ops_api=False,
        )
        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_json({"status": {"warnings": ["THE_WARNING", "THE_WARNING_2"]}})

        with caplog.at_level(logging.WARNING):
            cmd.request()
            w_records = [
                record
                for record in caplog.records
                if record.levelno == logging.WARNING
                if "THE_WARNING" in record.msg
            ]
            assert len(w_records) == 2
        caplog.clear()

        ops_base_path = "/base_ops"
        devops_cmd = APICommander(
            api_endpoint=base_endpoint,
            path=ops_base_path,
            dev_ops_api=True,
        )
        httpserver.expect_oneshot_request(
            ops_base_path,
        ).respond_with_json({"status": {"warnings": ["THE_WARNING"]}})

        with caplog.at_level(logging.WARNING):
            devops_cmd.request()
            w_records = [
                record
                for record in caplog.records
                if record.levelno == logging.WARNING
                if "THE_WARNING" in record.msg
            ]
            assert len(w_records) == 0
        caplog.clear()

    @pytest.mark.describe("test of APICommander server warnings, async")
    async def test_apicommander_server_warnings_async(
        self,
        caplog: pytest.LogCaptureFixture,
        httpserver: HTTPServer,
    ) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            dev_ops_api=False,
        )
        httpserver.expect_oneshot_request(
            base_path,
        ).respond_with_json({"status": {"warnings": ["THE_WARNING", "THE_WARNING_2"]}})

        with caplog.at_level(logging.WARNING):
            await cmd.async_request()
            w_records = [
                record
                for record in caplog.records
                if record.levelno == logging.WARNING
                if "THE_WARNING" in record.msg
            ]
            assert len(w_records) == 2
        caplog.clear()

        ops_base_path = "/base_ops"
        devops_cmd = APICommander(
            api_endpoint=base_endpoint,
            path=ops_base_path,
            dev_ops_api=True,
        )
        httpserver.expect_oneshot_request(
            ops_base_path,
        ).respond_with_json({"status": {"warnings": ["THE_WARNING"]}})

        with caplog.at_level(logging.WARNING):
            await devops_cmd.async_request()
            w_records = [
                record
                for record in caplog.records
                if record.levelno == logging.WARNING
                if "THE_WARNING" in record.msg
            ]
            assert len(w_records) == 0
        caplog.clear()



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import sys
from typing import Any, Iterable

from astrapy.info import EmbeddingProviderParameter, VectorServiceOptions
from astrapy.settings.defaults import (
    EMBEDDING_HEADER_API_KEY,
    EMBEDDING_HEADER_AWS_ACCESS_ID,
    EMBEDDING_HEADER_AWS_SECRET_ID,
)

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from live_provider_info import live_provider_info

alphanum = set("qwertyuiopasdfghjklzxcvbnmQWERTYUIOPASDFGHJKLZXCVBNM1234567890")


DEFAULT_TEST_ASSETS = {
    "samples": [
        (
            "car",
            "My car is fast and red, it's awesome.",
        ),
        (
            "landscape",
            "The landscape shows spiky, rocky mountaintops and a river nearby...",
        ),
        (
            "motorbike",
            "Your motorbike is certainly the best I have seen!",
        ),
    ],
    "probe": {
        "text": "Driving a red car is a joy.",
        "expected": ["car", "motorbike"],
    },
}

CODE_TEST_ASSETS = {
    "samples": [
        (
            "py_loop",
            "import os; for i in range(10): print(i + os.getenv('env'))",
        ),
        (
            "java",
            (
                "if (sortClause == null || sortClause.sortExpressions()."
                "isEmpty()){return Uni.createFrom().item(true);}"
            ),
        ),
        (
            "py_while",
            (
                "import json; val=json.load(open('limit.json')); "
                "v=0; while v<val: pprint(v); v = v + 1"
            ),
        ),
    ],
    "probe": {
        "text": "for(q in [10, 11, 12]): print(f'q={q}')",
        "expected": ["py_loop", "py_while"],
    },
}

TEST_ASSETS_MAP = {
    ("jinaAI", "jina-embeddings-v2-base-code"): CODE_TEST_ASSETS,
    ("voyageAI", "voyage-code-2"): CODE_TEST_ASSETS,
}

USE_INSERT_ONE_MAP: dict[tuple[str, str], bool] = {
    # ("upstageAI", "solar-1-mini-embedding"): True,
}

SECRET_NAME_ROOT_MAP = {
    "azureOpenAI": "AZURE_OPENAI",
    "bedrock": "BEDROCK",
    "cohere": "COHERE",
    "huggingface": "HUGGINGFACE",
    "huggingfaceDedicated": "HUGGINGFACEDED",
    "jinaAI": "JINAAI",
    "mistral": "MISTRAL",
    "nvidia": "NVIDIA",
    "openai": "OPENAI",
    "upstageAI": "UPSTAGE",
    "voyageAI": "VOYAGEAI",
}

# Any "model tag" listed here is skipped no matter what.
# (useful for models already from the findEmbProvs but not stable/working yet).
EXCLUDED_MODEL_QUADRUPLES = {
    ("jinaAI", "jina-embeddings-v3", "HEADER", "0"),
}

# this is a way to suppress/limit certain combinations of
# "full param" testing from the start. If even one of the optional params
# in a test model is PARAM_SKIP_MARKER, the combination is not emitted at all.
PARAM_SKIP_MARKER = "__SKIP_ME__"

PARAMETER_VALUE_MAP = {
    ("azureOpenAI", "text-embedding-3-large", "deploymentId"): os.environ[
        "AZURE_OPENAI_DEPLOY_ID_EMB3LARGE"
    ],
    ("azureOpenAI", "text-embedding-3-large", "resourceName"): os.environ[
        "AZURE_OPENAI_RESNAME_EMB3LARGE"
    ],
    ("azureOpenAI", "text-embedding-3-small", "deploymentId"): os.environ[
        "AZURE_OPENAI_DEPLOY_ID_EMB3SMALL"
    ],
    ("azureOpenAI", "text-embedding-3-small", "resourceName"): os.environ[
        "AZURE_OPENAI_RESNAME_EMB3SMALL"
    ],
    ("azureOpenAI", "text-embedding-ada-002", "deploymentId"): os.environ[
        "AZURE_OPENAI_DEPLOY_ID_ADA2"
    ],
    ("azureOpenAI", "text-embedding-ada-002", "resourceName"): os.environ[
        "AZURE_OPENAI_RESNAME_ADA2"
    ],
    ("jinaAI", "jina-embeddings-v3", "late_chunking"): True,
    ("jinaAI", "jina-embeddings-v3", "task"): "text-matching",
    ("voyageAI", "voyage-2", "autoTruncate"): True,
    ("voyageAI", "voyage-code-2", "autoTruncate"): True,
    ("voyageAI", "voyage-finance-2", "autoTruncate"): True,
    ("voyageAI", "voyage-large-2", "autoTruncate"): True,
    ("voyageAI", "voyage-large-2-instruct", "autoTruncate"): True,
    ("voyageAI", "voyage-law-2", "autoTruncate"): True,
    ("voyageAI", "voyage-multilingual-2", "autoTruncate"): True,
    #
    ("huggingfaceDedicated", "endpoint-defined-model", "endpointName"): os.environ[
        "HUGGINGFACEDED_ENDPOINTNAME"
    ],
    ("huggingfaceDedicated", "endpoint-defined-model", "regionName"): os.environ[
        "HUGGINGFACEDED_REGIONNAME"
    ],
    ("huggingfaceDedicated", "endpoint-defined-model", "cloudName"): os.environ[
        "HUGGINGFACEDED_CLOUDNAME"
    ],
    #
    ("openai", "text-embedding-3-large", "organizationId"): os.environ[
        "OPENAI_ORGANIZATION_ID"
    ],
    ("openai", "text-embedding-3-large", "projectId"): os.environ["OPENAI_PROJECT_ID"],
    ("openai", "text-embedding-3-small", "organizationId"): os.environ[
        "OPENAI_ORGANIZATION_ID"
    ],
    ("openai", "text-embedding-3-small", "projectId"): os.environ["OPENAI_PROJECT_ID"],
    ("openai", "text-embedding-ada-002", "organizationId"): os.environ[
        "OPENAI_ORGANIZATION_ID"
    ],
    ("openai", "text-embedding-ada-002", "projectId"): os.environ["OPENAI_PROJECT_ID"],
    #
    ("bedrock", "amazon.titan-embed-text-v1", "region"): os.environ["BEDROCK_REGION"],
    ("bedrock", "amazon.titan-embed-text-v2:0", "region"): os.environ["BEDROCK_REGION"],
}

# this is ad-hoc for HF dedicated. Models here, though "optional" dimension,
# do not undergo the f/0 optional dimension because of that, rather have
# a forced fixed, provided dimension.
FORCE_DIMENSION_MAP = {
    ("huggingfaceDedicated", "endpoint-defined-model"): int(
        os.environ["HUGGINGFACEDED_DIMENSION"]
    ),
}


def live_test_models() -> Iterable[dict[str, Any]]:
    def _from_validation(pspec: EmbeddingProviderParameter) -> int:
        assert pspec.parameter_type.upper() == "NUMBER"
        if "numericRange" in pspec.validation:
            m0: int = pspec.validation["numericRange"][0]
            m1: int = pspec.validation["numericRange"][1]
            return (m0 + m1) // 2
        elif "options" in pspec.validation:
            options: list[int] = pspec.validation["options"]
            if len(options) > 1:
                return options[1]
            else:
                return options[0]
        else:
            raise ValueError("unsupported pspec")

    def _collapse(longt: str) -> str:
        if len(longt) <= 40:
            return longt
        else:
            return f"{longt[:30]}_{longt[-5:]}"

    # generate the full list of models based on the live provider endpoint
    provider_info = live_provider_info()
    for provider_name, provider_desc in sorted(
        provider_info.embedding_providers.items()
    ):
        for model in provider_desc.models:
            for auth_type_name, auth_type_desc in sorted(
                provider_desc.supported_authentication.items()
            ):
                if auth_type_desc.enabled:
                    # test assumptions on auth type
                    if auth_type_name == "NONE":
                        assert auth_type_desc.tokens == []
                    elif auth_type_name == "HEADER":
                        header_names_lower = tuple(
                            sorted(t.accepted.lower() for t in auth_type_desc.tokens)
                        )
                        assert header_names_lower in {
                            (EMBEDDING_HEADER_API_KEY.lower(),),
                            (
                                EMBEDDING_HEADER_AWS_ACCESS_ID.lower(),
                                EMBEDDING_HEADER_AWS_SECRET_ID.lower(),
                            ),
                        }
                    elif auth_type_name == "SHARED_SECRET":
                        authkey_names = tuple(
                            sorted(t.accepted for t in auth_type_desc.tokens)
                        )
                        assert authkey_names in {
                            ("providerKey",),
                            ("accessId", "secretKey"),
                        }
                    else:
                        raise ValueError("Unknown auth type")

                    # params
                    collated_params = provider_desc.parameters + model.parameters
                    all_nond_params = [
                        param
                        for param in collated_params
                        if param.name != "vectorDimension"
                    ]
                    required_nond_params = {
                        param.name for param in all_nond_params if param.required
                    }
                    optional_nond_params = {
                        param.name for param in all_nond_params if not param.required
                    }
                    #
                    d_params = [
                        param
                        for param in collated_params
                        if param.name == "vectorDimension"
                    ]
                    if d_params:
                        d_param = d_params[0]
                        if (provider_name, model.name) in FORCE_DIMENSION_MAP:
                            optional_dimension = False
                            dimension = FORCE_DIMENSION_MAP[(provider_name, model.name)]
                        elif d_param.default_value is not None:
                            optional_dimension = True
                            assert model.vector_dimension is None
                            dimension = _from_validation(d_param)
                        else:
                            optional_dimension = False
                            assert model.vector_dimension is None
                            dimension = _from_validation(d_param)
                    else:
                        optional_dimension = False
                        assert model.vector_dimension is not None
                        assert model.vector_dimension > 0
                        dimension = model.vector_dimension

                    model_parameters = {
                        param_name: PARAMETER_VALUE_MAP[
                            (provider_name, model.name, param_name)
                        ]
                        for param_name in required_nond_params
                    }
                    optional_model_parameters = {
                        param_name: PARAMETER_VALUE_MAP[
                            (provider_name, model.name, param_name)
                        ]
                        for param_name in optional_nond_params
                    }

                    minimal_model_quadruple = (
                        provider_name,
                        model.name,
                        auth_type_name,
                        "0",
                    )
                    if minimal_model_quadruple not in EXCLUDED_MODEL_QUADRUPLES:
                        if optional_dimension or optional_nond_params != set():
                            # we issue a minimal-params version
                            model_tag_0 = (
                                f"{provider_name}/{model.name}/{auth_type_name}/0"
                            )
                            this_minimal_model = {
                                "model_tag": model_tag_0,
                                "simple_tag": _collapse(
                                    "".join(c for c in model_tag_0 if c in alphanum)
                                ),
                                "auth_type_name": auth_type_name,
                                "auth_type_tokens": auth_type_desc.tokens,
                                "secret_tag": SECRET_NAME_ROOT_MAP[provider_name],
                                "test_assets": TEST_ASSETS_MAP.get(
                                    (provider_name, model.name), DEFAULT_TEST_ASSETS
                                ),
                                "use_insert_one": USE_INSERT_ONE_MAP.get(
                                    (provider_name, model.name), False
                                ),
                                "service_options": VectorServiceOptions(
                                    provider=provider_name,
                                    model_name=model.name,
                                    parameters=model_parameters,
                                ),
                            }
                            yield this_minimal_model

                    # and in any case we issue a 'full-spec' one ...
                    # ... unless explicitly marked as skipped
                    if all(
                        v != PARAM_SKIP_MARKER
                        for v in optional_model_parameters.values()
                    ):
                        root_model = {
                            "auth_type_name": auth_type_name,
                            "auth_type_tokens": auth_type_desc.tokens,
                            "dimension": dimension,
                            "secret_tag": SECRET_NAME_ROOT_MAP[provider_name],
                            "test_assets": TEST_ASSETS_MAP.get(
                                (provider_name, model.name), DEFAULT_TEST_ASSETS
                            ),
                            "use_insert_one": USE_INSERT_ONE_MAP.get(
                                (provider_name, model.name), False
                            ),
                        }

                        model_tag_f = f"{provider_name}/{model.name}/{auth_type_name}/f"
                        full_model_quadruple = (
                            provider_name,
                            model.name,
                            auth_type_name,
                            "f",
                        )
                        if full_model_quadruple not in EXCLUDED_MODEL_QUADRUPLES:
                            this_model = {
                                "model_tag": model_tag_f,
                                "simple_tag": _collapse(
                                    "".join(c for c in model_tag_f if c in alphanum)
                                ),
                                "service_options": VectorServiceOptions(
                                    provider=provider_name,
                                    model_name=model.name,
                                    parameters={
                                        **model_parameters,
                                        **optional_model_parameters,
                                    },
                                ),
                                **root_model,
                            }
                            yield this_model



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os

import pytest

from astrapy.api_options import APIOptions
from astrapy.data_types import DataAPIVector
from astrapy.exceptions import TableInsertManyException

from ..conftest import IS_ASTRA_DB, VECTORIZE_TEXTS, DefaultAsyncTable

HEADER_EMBEDDING_API_KEY_OPENAI = os.environ.get("HEADER_EMBEDDING_API_KEY_OPENAI")


class TestTableVectorizeAsync:
    @pytest.mark.skipif(
        HEADER_EMBEDDING_API_KEY_OPENAI is None,
        reason="No HEADER_EMBEDDING_API_KEY_OPENAI detected",
    )
    @pytest.mark.describe("test of basic table vectorize with key via header, async")
    async def test_table_vectorize_header_async(
        self,
        async_empty_table_vectorize: DefaultAsyncTable,
    ) -> None:
        aauthenticated_table = async_empty_table_vectorize.with_options(
            api_options=APIOptions(
                embedding_api_key=HEADER_EMBEDDING_API_KEY_OPENAI or "",
            ),
        )
        i_vector = DataAPIVector([0.01] * 64)
        await aauthenticated_table.insert_one(
            {"p_text": "v", "p_vector": i_vector},
        )
        await aauthenticated_table.insert_one(
            {"p_text": "t", "p_vector": "This is a text example."},
        )
        await aauthenticated_table.insert_one(
            {"p_text": "u"},
        )
        await aauthenticated_table.insert_one(
            {"p_text": "n", "p_vector": None},
        )
        v_row = await aauthenticated_table.find_one(filter={"p_text": "v"})
        t_row = await aauthenticated_table.find_one(filter={"p_text": "t"})
        u_row = await aauthenticated_table.find_one(filter={"p_text": "u"})
        n_row = await aauthenticated_table.find_one(filter={"p_text": "n"})
        assert v_row is not None
        assert t_row is not None
        assert u_row is not None
        assert n_row is not None
        assert isinstance(v_row["p_vector"], DataAPIVector)
        assert v_row["p_vector"] == i_vector
        assert u_row["p_vector"] is None
        assert n_row["p_vector"] is None

        vec_row = await aauthenticated_table.find_one(
            sort={"p_vector": "This is one text example."}
        )
        assert vec_row == t_row

        await aauthenticated_table.delete_many({})
        await aauthenticated_table.insert_many(
            [
                {"p_text": f"t_{text_i}", "p_vector": text}
                for text_i, text in enumerate(VECTORIZE_TEXTS)
            ],
            chunk_size=min(20, 1 + len(VECTORIZE_TEXTS) // 3),
        )
        all_rows = await aauthenticated_table.find({}).to_list()
        assert len(all_rows) == len(VECTORIZE_TEXTS)
        assert all(isinstance(row["p_vector"], DataAPIVector) for row in all_rows)

        # include sort vector and include similarity
        cur_isv = aauthenticated_table.find(
            sort={"p_vector": "This is a fact."},
            include_sort_vector=True,
            include_similarity=True,
        )
        # TODO: restore this check once #1949 is resolved
        # assert isinstance(await cur_isv.get_sort_vector(), DataAPIVector)
        match0 = await cur_isv.__anext__()
        assert match0 is not None
        # TODO: restore this check once #1949 is resolved
        # assert isinstance(match0["$similarity"], float)

    @pytest.mark.skipif(
        HEADER_EMBEDDING_API_KEY_OPENAI is None,
        reason="No HEADER_EMBEDDING_API_KEY_OPENAI detected",
    )
    @pytest.mark.describe(
        "test of vectorize-based table insert many grand failures, async"
    )
    async def test_table_insertmany_grandfailure_vectorize_async(
        self,
        async_empty_table_vectorize: DefaultAsyncTable,
    ) -> None:
        authenticated_atable = async_empty_table_vectorize.with_options(
            api_options=APIOptions(
                embedding_api_key=HEADER_EMBEDDING_API_KEY_OPENAI or "",
            ),
        )
        await authenticated_atable.insert_many([{"p_text": "0", "p_vector": "Text."}])

        bad_atable = authenticated_atable.with_options(embedding_api_key="BadKey")

        err: TableInsertManyException | None = None
        try:
            await bad_atable.insert_many(
                [{"p_text": "1"}, {"p_text": "2", "p_vector": "Text."}]
            )
        except TableInsertManyException as e:
            err = e
        assert err is not None
        assert err.inserted_ids == []

    @pytest.mark.skipif(
        not IS_ASTRA_DB,
        reason="KMS vectorize is available only on Astra DB",
    )
    @pytest.mark.describe(
        "test of basic table vectorize with key via shared secret, async"
    )
    async def test_table_vectorize_shared_secret_async(
        self,
        async_empty_table_kms_vectorize: DefaultAsyncTable,
    ) -> None:
        i_vector = DataAPIVector([0.01] * 64)
        await async_empty_table_kms_vectorize.insert_one(
            {"p_text": "v", "p_vector": i_vector},
        )
        await async_empty_table_kms_vectorize.insert_one(
            {"p_text": "t", "p_vector": "This is a text example."},
        )
        await async_empty_table_kms_vectorize.insert_one(
            {"p_text": "u"},
        )
        await async_empty_table_kms_vectorize.insert_one(
            {"p_text": "n", "p_vector": None},
        )
        v_row = await async_empty_table_kms_vectorize.find_one(filter={"p_text": "v"})
        t_row = await async_empty_table_kms_vectorize.find_one(filter={"p_text": "t"})
        u_row = await async_empty_table_kms_vectorize.find_one(filter={"p_text": "u"})
        n_row = await async_empty_table_kms_vectorize.find_one(filter={"p_text": "n"})
        assert v_row is not None
        assert t_row is not None
        assert u_row is not None
        assert n_row is not None
        assert isinstance(v_row["p_vector"], DataAPIVector)
        assert v_row["p_vector"] == i_vector
        assert u_row["p_vector"] is None
        assert n_row["p_vector"] is None

        vec_row = await async_empty_table_kms_vectorize.find_one(
            sort={"p_vector": "This is one text example."}
        )
        assert vec_row == t_row

        await async_empty_table_kms_vectorize.delete_many({})
        await async_empty_table_kms_vectorize.insert_many(
            [
                {"p_text": f"t_{text_i}", "p_vector": text}
                for text_i, text in enumerate(VECTORIZE_TEXTS)
            ],
            chunk_size=min(20, 1 + len(VECTORIZE_TEXTS) // 3),
        )
        all_rows = await async_empty_table_kms_vectorize.find({}).to_list()
        assert len(all_rows) == len(VECTORIZE_TEXTS)
        assert all(isinstance(row["p_vector"], DataAPIVector) for row in all_rows)

        # include sort vector and include similarity
        cur_isv = async_empty_table_kms_vectorize.find(
            sort={"p_vector": "This is a fact."},
            include_sort_vector=True,
            include_similarity=True,
        )

        # TODO: restore this check once #1949 is resolved
        # assert isinstance(await cur_isv.get_sort_vector(), DataAPIVector)
        match0 = await cur_isv.__anext__()
        assert match0 is not None
        # TODO: restore this check once #1949 is resolved
        # assert isinstance(match0["$similarity"], float)

    @pytest.mark.skipif(
        HEADER_EMBEDDING_API_KEY_OPENAI is None,
        reason="No HEADER_EMBEDDING_API_KEY_OPENAI detected",
    )
    @pytest.mark.describe("test of multiple-vectorize table usage via header, async")
    async def test_table_multiplevectorize_header_async(
        self,
        async_empty_table_multiplevectorize: DefaultAsyncTable,
    ) -> None:
        atable = async_empty_table_multiplevectorize

        await atable.insert_one(
            {
                "id": "A",
                "p_text": "p_text of A",
                "p_vector_small": "a sentence on mountains",
                "p_vector_large": "telephones and electricity?",
            }
        )

        await atable.insert_many(
            [
                {
                    "id": "B",
                    "p_text": "p_text of B",
                    "p_vector_small": "rivers, and has only the bee",
                },
                {
                    "id": "C",
                    "p_text": "p_text of C",
                    "p_vector_large": "clouds, and has only the sea",
                },
            ]
        )

        reg_f_rows = await atable.find(projection={"id": True}).to_list()
        assert {row["id"] for row in reg_f_rows} == {"A", "B", "C"}

        ann_f_small = await atable.find(
            sort={"p_vector_small": "flowing water and honey-producing animals"},
            projection={"id": True},
        ).to_list()
        assert ann_f_small != []
        assert ann_f_small[0]["id"] == "B"

        ann_f1_small = await atable.find_one(
            sort={"p_vector_small": "the utterance about orography"},
            projection={"id": True},
        )
        assert ann_f1_small is not None
        assert ann_f1_small["id"] == "A"

        await atable.insert_one(
            {
                "id": "Z",
                "p_text": "p_text of Z",
                "p_vector_small": "telephones and electricity?",
                "p_vector_large": "a sentence on mountains",
            }
        )

        f1_ele_doc_sm = await atable.find_one(
            sort={"p_vector_small": "telephones and electricity"}
        )
        f1_ele_doc_lg = await atable.find_one(
            sort={"p_vector_large": "telephones and electricity"}
        )
        assert f1_ele_doc_sm is not None
        # assert f1_ele_doc_sm["id"] == "Z"  # suspended (delays in SAI, nondeterm.)
        assert f1_ele_doc_lg is not None
        # assert f1_ele_doc_lg["id"] == "A"  # suspended (delays in SAI, nondeterm.)

        await atable.update_one(
            {"id": "A"},
            update={
                "$set": {
                    "p_text": "the new p_text for A",
                    "p_vector_small": "glaciers among the woods.",
                },
            },
        )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for the parsing of API endpoints and related
"""

from __future__ import annotations

import json
import time
from typing import Any, Callable

import pytest
import werkzeug
from pytest_httpserver import HTTPServer

from astrapy import DataAPIClient
from astrapy.api_options import APIOptions, TimeoutOptions
from astrapy.exceptions import (
    DataAPITimeoutException,
    DevOpsAPITimeoutException,
    _TimeoutContext,
)
from astrapy.utils.api_commander import APICommander
from astrapy.utils.request_tools import HttpMethod

SLEEPER_TIME_MS = 500
TIMEOUT_PARAM_MS = 100


def response_sleeper(request: werkzeug.Request) -> werkzeug.Response:
    time.sleep(SLEEPER_TIME_MS / 1000)
    return werkzeug.Response()


def delayed_response_handler(
    delay_ms: int, response_json: dict[str, Any]
) -> Callable[[werkzeug.Request], werkzeug.Response]:
    def _response_sleeper(request: werkzeug.Request) -> werkzeug.Response:
        time.sleep(delay_ms / 1000)
        return werkzeug.Response(json.dumps(response_json))

    return _response_sleeper


@pytest.mark.skip("Suppressed. The HTTPServer plays tricks here sometimes")
class TestTimeouts:
    @pytest.mark.describe("test of APICommander timeout, sync")
    def test_apicommander_timeout_sync(self, httpserver: HTTPServer) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
        )

        httpserver.expect_oneshot_request(
            base_path,
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DataAPITimeoutException):
            cmd.request(timeout_context=_TimeoutContext(request_ms=TIMEOUT_PARAM_MS))

    @pytest.mark.describe("test of APICommander timeout, async")
    async def test_apicommander_timeout_async(self, httpserver: HTTPServer) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
        )

        httpserver.expect_oneshot_request(
            base_path,
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DataAPITimeoutException):
            await cmd.async_request(
                timeout_context=_TimeoutContext(request_ms=TIMEOUT_PARAM_MS)
            )

    @pytest.mark.describe("test of APICommander timeout DevOps, sync")
    def test_apicommander_timeout_devops_sync(self, httpserver: HTTPServer) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            dev_ops_api=True,
        )

        httpserver.expect_oneshot_request(
            base_path,
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DevOpsAPITimeoutException):
            cmd.request(timeout_context=_TimeoutContext(request_ms=TIMEOUT_PARAM_MS))

    @pytest.mark.describe("test of APICommander timeout DevOps, async")
    async def test_apicommander_timeout_devops_async(
        self, httpserver: HTTPServer
    ) -> None:
        base_endpoint = httpserver.url_for("/")
        base_path = "/base"
        cmd = APICommander(
            api_endpoint=base_endpoint,
            path=base_path,
            dev_ops_api=True,
        )

        httpserver.expect_oneshot_request(
            base_path,
            method=HttpMethod.POST,
        ).respond_with_handler(response_sleeper)
        with pytest.raises(DevOpsAPITimeoutException):
            await cmd.async_request(
                timeout_context=_TimeoutContext(request_ms=TIMEOUT_PARAM_MS)
            )

    @pytest.mark.describe(
        "test of timeout occurring, zero and nonzero, for Collection class, sync"
    )
    def test_collection_timeout_occurring_sync(self, httpserver: HTTPServer) -> None:
        root_endpoint = httpserver.url_for("/")
        client = DataAPIClient(environment="other")
        database = client.get_database(root_endpoint, keyspace="xkeyspace")
        # S_L = short timeout for request; long timeout for general-method
        collection_S_L = database.get_collection(
            "xcollt",
            spawn_api_options=APIOptions(
                timeout_options=TimeoutOptions(
                    request_timeout_ms=1,
                    general_method_timeout_ms=10000,
                ),
            ),
        )
        collection_L_S = database.get_collection(
            "xcollt",
            spawn_api_options=APIOptions(
                timeout_options=TimeoutOptions(
                    request_timeout_ms=10000,
                    general_method_timeout_ms=1,
                ),
            ),
        )
        expected_url = "/v1/xkeyspace/xcollt"

        # no method arg, the short request_ms makes it timeout
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        with pytest.raises(DataAPITimeoutException):
            collection_S_L.delete_many({})

        # no method arg, the short generalmethod makes it timeout
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        with pytest.raises(DataAPITimeoutException):
            collection_L_S.delete_many({})

        # this avoids dangling timing-out response to pollute next test:
        httpserver.stop()  # type: ignore[no-untyped-call]
        httpserver.start()  # type: ignore[no-untyped-call]

    @pytest.mark.describe(
        "test of timeout suppression, zero and nonzero, for Collection class, sync"
    )
    def test_collection_timeout_suppression_sync(self, httpserver: HTTPServer) -> None:
        # various ways to raise or disable timeouts and have requests succeed on time
        root_endpoint = httpserver.url_for("/")
        client = DataAPIClient(environment="other")
        database = client.get_database(root_endpoint, keyspace="xkeyspace")
        # S_L = short timeout for request; long timeout for general-method
        collection_S_L = database.get_collection(
            "xcoll",
            spawn_api_options=APIOptions(
                timeout_options=TimeoutOptions(
                    request_timeout_ms=1,
                    general_method_timeout_ms=10000,
                ),
            ),
        )
        collection_L_S = database.get_collection(
            "xcoll",
            spawn_api_options=APIOptions(
                timeout_options=TimeoutOptions(
                    request_timeout_ms=10000,
                    general_method_timeout_ms=1,
                ),
            ),
        )
        expected_url = "/v1/xkeyspace/xcoll"

        # remove the timeout with method reqtimeout override
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_SL_rq = collection_S_L.delete_many({}, request_timeout_ms=1500)
        assert dmr_SL_rq.deleted_count == 12

        # remove the timeout with method genmeth-timeout override
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_LS_rq = collection_L_S.delete_many({}, general_method_timeout_ms=1500)
        assert dmr_LS_rq.deleted_count == 12

        # remove the timeout completely with a zero per-method per-req timeout
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_SL_zrq = collection_S_L.delete_many({}, request_timeout_ms=0)
        assert dmr_SL_zrq.deleted_count == 12

        # remove the timeout completely with a zero per-method genmeth-timeout override
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_LS_zrq = collection_L_S.delete_many({}, general_method_timeout_ms=0)
        assert dmr_LS_zrq.deleted_count == 12

        # remove the timeout completely: a zero per-method 'general_method_timeout_ms'
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_LS_zgrq = collection_L_S.delete_many({}, general_method_timeout_ms=0)
        assert dmr_LS_zgrq.deleted_count == 12

    @pytest.mark.describe(
        "test of timeout occurring, zero and nonzero, for Collection class, async"
    )
    async def test_collection_timeout_occurring_async(
        self, httpserver: HTTPServer
    ) -> None:
        root_endpoint = httpserver.url_for("/")
        client = DataAPIClient(environment="other")
        database = client.get_database(root_endpoint, keyspace="xkeyspace")
        # S_L = short timeout for request; long timeout for general-method
        acollection_S_L = database.to_async().get_collection(
            "xcollt",
            spawn_api_options=APIOptions(
                timeout_options=TimeoutOptions(
                    request_timeout_ms=1,
                    general_method_timeout_ms=10000,
                ),
            ),
        )
        acollection_L_S = database.to_async().get_collection(
            "xcollt",
            spawn_api_options=APIOptions(
                timeout_options=TimeoutOptions(
                    request_timeout_ms=10000,
                    general_method_timeout_ms=1,
                ),
            ),
        )
        expected_url = "/v1/xkeyspace/xcollt"

        # no method arg, the short request_ms makes it timeout
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        with pytest.raises(DataAPITimeoutException):
            await acollection_S_L.delete_many({})

        # no method arg, the short generalmethod makes it timeout
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        with pytest.raises(DataAPITimeoutException):
            await acollection_L_S.delete_many({})

        # this avoids dangling timing-out response to pollute next test:
        httpserver.stop()  # type: ignore[no-untyped-call]
        httpserver.start()  # type: ignore[no-untyped-call]

    @pytest.mark.describe(
        "test of timeout suppression, zero and nonzero, for Collection class, async"
    )
    async def test_collection_timeout_suppression_async(
        self, httpserver: HTTPServer
    ) -> None:
        # various ways to raise or disable timeouts and have requests succeed on time
        root_endpoint = httpserver.url_for("/")
        client = DataAPIClient(environment="other")
        database = client.get_database(root_endpoint, keyspace="xkeyspace")
        # S_L = short timeout for request; long timeout for general-method
        acollection_S_L = database.to_async().get_collection(
            "xcoll",
            spawn_api_options=APIOptions(
                timeout_options=TimeoutOptions(
                    request_timeout_ms=1,
                    general_method_timeout_ms=10000,
                ),
            ),
        )
        acollection_L_S = database.to_async().get_collection(
            "xcoll",
            spawn_api_options=APIOptions(
                timeout_options=TimeoutOptions(
                    request_timeout_ms=10000,
                    general_method_timeout_ms=1,
                ),
            ),
        )
        expected_url = "/v1/xkeyspace/xcoll"

        # remove the timeout with method reqtimeout override
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_SL_rq = await acollection_S_L.delete_many({}, request_timeout_ms=1500)
        assert dmr_SL_rq.deleted_count == 12

        # remove the timeout with method genmeth-timeout override
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_LS_rq = await acollection_L_S.delete_many(
            {}, general_method_timeout_ms=1500
        )
        assert dmr_LS_rq.deleted_count == 12

        # remove the timeout completely with a zero per-method per-req timeout
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_SL_zrq = await acollection_S_L.delete_many({}, request_timeout_ms=0)
        assert dmr_SL_zrq.deleted_count == 12

        # remove the timeout completely with a zero per-method genmeth-timeout override
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_LS_zrq = await acollection_L_S.delete_many({}, general_method_timeout_ms=0)
        assert dmr_LS_zrq.deleted_count == 12

        # remove the timeout completely: a zero per-method 'general_method_timeout_ms'
        httpserver.expect_oneshot_request(
            expected_url,
            method="POST",
        ).respond_with_handler(
            delayed_response_handler(
                delay_ms=1200,
                response_json={
                    "status": {
                        "deletedCount": 12,
                        "moreData": False,
                    },
                },
            )
        )
        dmr_LS_zgrq = await acollection_L_S.delete_many({}, general_method_timeout_ms=0)
        assert dmr_LS_zgrq.deleted_count == 12



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from astrapy import DataAPIClient

from ..conftest import IS_ASTRA_DB, DataAPICredentials


class TestFindRegionsSync:
    @pytest.mark.skipif(not IS_ASTRA_DB, reason="This test requires Astra DB")
    @pytest.mark.describe("test of find_available_regions, sync")
    def test_findavailableregions_sync(
        self,
        client: DataAPIClient,
        data_api_credentials_kwargs: DataAPICredentials,
    ) -> None:
        admin = client.get_admin(token=data_api_credentials_kwargs["token"])

        ar0 = admin.find_available_regions()
        art = admin.find_available_regions(only_org_enabled_regions=True)
        arf = admin.find_available_regions(only_org_enabled_regions=False)

        assert ar0 == art
        assert len(arf) >= len(art)
        assert all(reg in arf for reg in art)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import pytest

from ..conftest import DefaultTable


class TestTableSpecialShortformTypesInListTables:
    @pytest.mark.describe(
        "Test listTables returns the short form for key/valueType primitive fields"
    )
    def test_table_special_shortform_types_listtables(
        self,
        sync_table_all_returns: DefaultTable,
    ) -> None:
        """
        Temporary test added as a guard to ensure short-form is used
        for primitive types when inside collection columns. Once clients 2.1
        are out for some time, this can be lifted and the newer, long-form-preferring
        Data API can be deployed.
        """

        raw_response = sync_table_all_returns.database.command(
            {"listTables": {"options": {"explain": True}}}
        )
        the_table_s = [
            tb_dict
            for tb_dict in raw_response["status"]["tables"]
            if tb_dict["name"] == sync_table_all_returns.name
        ]
        assert the_table_s != []

        for col_name in {"p_list_int", "p_map_text_text", "p_set_int"}:
            col_def = the_table_s[0]["definition"]["columns"][col_name]
            if "valueType" in col_def:
                assert isinstance(col_def["valueType"], str)
            if "keyType" in col_def:
                assert isinstance(col_def["keyType"], str)

    @pytest.mark.describe(
        "Test projectionSchema has the short form for key/valueType primitive fields"
    )
    def test_table_special_shortform_types_projectionschema(
        self,
        sync_table_all_returns: DefaultTable,
    ) -> None:
        """
        Temporary test added as a guard to ensure short-form is used
        for primitive types when inside collection columns. Once clients 2.1
        are out for some time, this can be lifted and the newer, long-form-preferring
        Data API can be deployed.
        """

        raw_response = sync_table_all_returns.command(
            {"findOne": {"filter": {"p_ascii": "a", "p_bigint": 1}}}
        )
        proj_schema = raw_response["status"]["projectionSchema"]

        for col_name in {"p_list_int", "p_map_text_text", "p_set_int"}:
            col_def = proj_schema[col_name]
            if "valueType" in col_def:
                assert isinstance(col_def["valueType"], str)
            if "keyType" in col_def:
                assert isinstance(col_def["keyType"], str)



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any

import pytest

from astrapy import Database
from astrapy.data_types import DataAPIVector
from astrapy.exceptions import CollectionInsertManyException, DataAPIResponseException
from astrapy.info import CollectionDefinition

from ..conftest import (
    HEADER_EMBEDDING_API_KEY_OPENAI,
    DefaultCollection,
)


@pytest.mark.skipif(
    HEADER_EMBEDDING_API_KEY_OPENAI is None,
    reason="No HEADER_EMBEDDING_API_KEY_OPENAI credential",
)
class TestCollectionVectorizeMethodsSync:
    @pytest.mark.describe("test of vectorize in collection methods, sync")
    def test_collection_methods_vectorize_sync(
        self,
        sync_empty_service_collection: DefaultCollection,
        service_collection_parameters: dict[str, Any],
    ) -> None:
        col = sync_empty_service_collection
        service_vector_dimension = service_collection_parameters["dimension"]

        col.insert_one({"t": "tower", "$vectorize": "How high is this tower?"})
        col.insert_one({"t": "vectorless"})
        col.insert_one({"t": "vectorful", "$vector": [0.01] * service_vector_dimension})

        col.insert_many(
            [
                {
                    "t": "guide",
                    "$vectorize": "This is the instructions manual. Read it!",
                },
                {
                    "t": "seeds",
                    "$vectorize": "Other plants rely on wind to propagate their seeds.",
                },
                {
                    "t": "dog",
                    "$vector": [0.01] * service_vector_dimension,
                },
                {
                    "t": "cat_novector",
                },
                {
                    "t": "spider",
                    "$vectorize": "The eye pattern is a primary criterion to the family.",
                },
            ],
        )

        doc = col.find_one(
            {},
            sort={"$vectorize": "This building is five storeys tall."},
            projection={"$vector": False},
        )
        assert doc is not None
        assert doc["t"] == "tower"

        docs = list(
            col.find(
                {},
                sort={"$vectorize": "This building is five storeys tall."},
                limit=2,
                projection={"$vector": False},
            )
        )
        assert docs[0]["t"] == "tower"

        rdoc = col.find_one_and_replace(
            {},
            {"t": "spider", "$vectorize": "Check out the eyes!"},
            sort={"$vectorize": "The disposition of the eyes tells much"},
            projection={"$vector": False},
        )
        assert rdoc is not None
        assert rdoc["t"] == "spider"

        r1res = col.replace_one(
            {},
            {"t": "spider", "$vectorize": "Look at how the eyes are placed"},
            sort={"$vectorize": "The disposition of the eyes tells much"},
        )
        assert r1res.update_info["nModified"] == 1

        udoc = col.find_one_and_update(
            {},
            {"$set": {"$vectorize": "Consider consulting the how-to"}},
            sort={"$vectorize": "Have a look at the user guide..."},
            projection={"$vector": False},
        )
        assert udoc is not None
        # TODO: re-enable after CNDB #14524 is merged/deployed
        # assert udoc["t"] == "guide"

        u1res = col.update_one(
            {},
            {"$set": {"$vectorize": "Know how to operate it before doing so."}},
            sort={"$vectorize": "Have a look at the user guide..."},
        )
        assert u1res.update_info["nModified"] == 1

        ddoc = col.find_one_and_delete(
            {},
            sort={"$vectorize": "Some trees have seeds that are dispersed in the air!"},
            projection={"$vector": False},
        )
        assert ddoc is not None
        # TODO: re-enable after CNDB #14524 is merged/deployed
        # assert ddoc["t"] == "seeds"

        d1res = col.delete_one(
            {},
            sort={"$vectorize": "yet another giant construction in this suburb."},
        )
        assert d1res.deleted_count == 1

    @pytest.mark.describe(
        "test of include_sort_vector in collection vectorize find, sync"
    )
    def test_collection_include_sort_vector_vectorize_find_sync(
        self,
        sync_empty_service_collection: DefaultCollection,
    ) -> None:
        # with empty collection
        q_text = "A sentence for searching."

        def _is_vector(v: Any) -> bool:
            if isinstance(v, list) and isinstance(v[0], float):
                return True
            if isinstance(v, DataAPIVector):
                return True
            return False

        for include_sv in [False, True]:
            for sort_cl_label in ["vze"]:
                sort_cl_e: dict[str, Any] = {"$vectorize": q_text}
                vec_expected = include_sv and sort_cl_label == "vze"
                # pristine iterator
                this_ite_1 = sync_empty_service_collection.find(
                    {}, sort=sort_cl_e, include_sort_vector=include_sv
                )
                if vec_expected:
                    assert _is_vector(this_ite_1.get_sort_vector())
                else:
                    assert this_ite_1.get_sort_vector() is None
                # after exhaustion with empty
                all_items_1 = list(this_ite_1)
                assert all_items_1 == []
                if vec_expected:
                    assert _is_vector(this_ite_1.get_sort_vector())
                else:
                    assert this_ite_1.get_sort_vector() is None
                # directly exhausted before calling get_sort_vector
                this_ite_2 = sync_empty_service_collection.find(
                    {}, sort=sort_cl_e, include_sort_vector=include_sv
                )
                all_items_2 = list(this_ite_2)
                assert all_items_2 == []
                if vec_expected:
                    assert _is_vector(this_ite_2.get_sort_vector())
                else:
                    assert this_ite_2.get_sort_vector() is None
        sync_empty_service_collection.insert_many(
            [
                {"seq": i, "$vectorize": f"This is sentence number {i}"}
                for i in range(10)
            ]
        )
        # with non-empty collection
        for include_sv in [False, True]:
            for sort_cl_label in ["vze"]:
                sort_cl_f: dict[str, Any] = {"$vectorize": q_text}
                vec_expected = include_sv and sort_cl_label == "vze"
                # pristine iterator
                this_ite_1 = sync_empty_service_collection.find(
                    {}, sort=sort_cl_f, include_sort_vector=include_sv
                )
                if vec_expected:
                    assert _is_vector(this_ite_1.get_sort_vector())
                else:
                    assert this_ite_1.get_sort_vector() is None
                # after consuming one item
                first_seqs = [
                    doc["seq"] for doc in [this_ite_1.__next__(), this_ite_1.__next__()]
                ]
                if vec_expected:
                    assert _is_vector(this_ite_1.get_sort_vector())
                else:
                    assert this_ite_1.get_sort_vector() is None
                # after exhaustion with the rest
                last_seqs = [doc["seq"] for doc in list(this_ite_1)]
                assert len(set(last_seqs + first_seqs)) == 10
                assert len(last_seqs + first_seqs) == 10
                if vec_expected:
                    assert _is_vector(this_ite_1.get_sort_vector())
                else:
                    assert this_ite_1.get_sort_vector() is None
                # directly exhausted before calling get_sort_vector
                this_ite_2 = sync_empty_service_collection.find(
                    {}, sort=sort_cl_f, include_sort_vector=include_sv
                )
                list(this_ite_2)
                if vec_expected:
                    assert _is_vector(this_ite_2.get_sort_vector())
                else:
                    assert this_ite_2.get_sort_vector() is None

    @pytest.mark.describe("test of vectorize-based insert many grand failures, sync")
    def test_collection_methods_grandfailure_vectorize_sync(
        self,
        sync_empty_service_collection: DefaultCollection,
    ) -> None:
        sync_empty_service_collection.insert_many([{"_id": "Z", "$vectorize": "Text."}])

        bad_collection = sync_empty_service_collection.with_options(
            embedding_api_key="BadKey",
        )
        err: CollectionInsertManyException | None = None
        try:
            bad_collection.insert_many(
                [{"_id": "A"}, {"_id": "B", "$vectorize": "Text."}]
            )
        except CollectionInsertManyException as e:
            err = e
        assert err is not None
        assert err.inserted_ids == []

    @pytest.mark.describe(
        "test of database create_collection dimension-mismatch failure, sync"
    )
    def test_database_create_collection_dimension_mismatch_failure_sync(
        self,
        sync_database: Database,
        service_collection_parameters: dict[str, Any],
    ) -> None:
        with pytest.raises(DataAPIResponseException):
            sync_database.create_collection(
                "collection_name",
                definition=(
                    CollectionDefinition.builder()
                    .set_vector_dimension(
                        service_collection_parameters["dimension"] + 10
                    )
                    .set_vector_service(
                        provider=service_collection_parameters["provider"],
                        model_name=service_collection_parameters["modelName"],
                    )
                    .build()
                ),
            )



# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import datetime
import decimal
import ipaddress
from typing import Any

import pytest

from astrapy.data.utils.extended_json_converters import (
    convert_ejson_binary_object_to_bytes,
)
from astrapy.data.utils.table_converters import (
    create_key_ktpostprocessor,
    create_row_tpostprocessor,
    preprocess_table_payload,
)
from astrapy.data.utils.table_types import ColumnType
from astrapy.data_types import (
    DataAPIDate,
    DataAPIDuration,
    DataAPIMap,
    DataAPISet,
    DataAPITime,
    DataAPITimestamp,
    DataAPIVector,
)
from astrapy.data_types.data_api_vector import bytes_to_floats
from astrapy.ids import UUID, ObjectId
from astrapy.info import ListTableDescriptor, TableScalarColumnTypeDescriptor
from astrapy.utils.api_options import SerdesOptions, defaultSerdesOptions

from ..conftest import _repaint_NaNs

TABLE_DESCRIPTION = {
    "name": "table_simple",
    "definition": {
        "columns": {
            "p_text": {"type": "text"},
            "p_boolean": {"type": "boolean"},
            "p_int": {"type": "int"},
            "p_float": {"type": "float"},
            "p_float_nan": {"type": "float"},
            "p_float_pinf": {"type": "float"},
            "p_float_minf": {"type": "float"},
            "p_blob": {"type": "blob"},
            "p_uuid": {"type": "uuid"},
            "p_decimal": {"type": "decimal"},
            "p_date": {"type": "date"},
            "p_duration": {"type": "duration"},
            "p_inet": {"type": "inet"},
            "p_time": {"type": "time"},
            "p_timestamp": {"type": "timestamp"},
            "p_timestamp_out_offset": {"type": "timestamp"},
            "p_list_int": {"type": "list", "valueType": "int"},
            "p_list_double": {"type": "list", "valueType": "double"},
            "p_set_ascii": {"type": "set", "valueType": "ascii"},
            "p_set_float": {"type": "set", "valueType": "float"},
            "p_map_text_float": {
                "type": "map",
                "valueType": "float",
                "keyType": "text",
            },
            "p_map_float_text": {
                "type": "map",
                "valueType": "text",
                "keyType": "float",
            },
            "somevector": {"type": "vector", "dimension": 3},
            "embeddings": {
                "type": "vector",
                "dimension": 2,
                "service": {
                    "provider": "openai",
                    "modelName": "text-embedding-3-small",
                },
            },
            "p_counter": {
                "type": "counter",
                "apiSupport": {
                    "createTable": False,
                    "insert": False,
                    "filter": True,
                    "read": True,
                    "cqlDefinition": "counter",
                },
            },
            "p_timeuuid": {
                "type": "timeuuid",
                "apiSupport": {
                    "createTable": False,
                    "insert": True,
                    "filter": True,
                    "read": True,
                    "cqlDefinition": "timeuuid",
                },
            },
        },
        "primaryKey": {"partitionBy": [], "partitionSort": {}},
    },
}

OUTPUT_ROW_TO_POSTPROCESS = {
    "p_text": "italy",
    "p_boolean": True,
    "p_int": 123,
    "p_float": 1.2,
    "p_float_nan": "NaN",
    "p_float_pinf": "Infinity",
    "p_float_minf": "-Infinity",
    "p_blob": {"$binary": "YWJjMTIz"},  # i.e. b"abc123"
    "p_uuid": "9c5b94b1-35ad-49bb-b118-8e8fc24abf80",
    "p_decimal": decimal.Decimal("123.456"),
    "p_date": "+11111-09-30",
    "p_duration": "P1M1DT1M0.000000001S",
    "p_inet": "10.1.1.2",
    "p_time": "12:34:56.78912",
    "p_timestamp": "2015-05-03T13:30:54.234Z",
    "p_timestamp_out_offset": "-0123-04-03T13:13:04.123+1:00",
    "p_list_int": [99, 100, 101],
    "p_list_double": [1.1, 2.2, "NaN", "Infinity", "-Infinity", 9.9],
    "p_set_ascii": ["a", "b", "c"],
    "p_set_float": [1.1, "NaN", "-Infinity", 9.9],
    "p_map_text_float": {"a": 0.1, "b": 0.2},
    "p_map_float_text": {1.1: "1-1", "NaN": "NANNN!"},
    "somevector": {"$binary": "PczMzb5MzM0+mZma"},  # i.e. [0.1,-0.2,0.3], not bitwise
    "embeddings": [0.1, -0.2, 0.3],
    "p_counter": 100,
    "p_timeuuid": "0de779c0-92e3-11ef-96a4-a745ae2c0a0b",
}

EXPECTED_POSTPROCESSED_ROW = {
    "p_text": "italy",
    "p_boolean": True,
    "p_int": 123,
    "p_float": 1.2,
    "p_float_nan": float("NaN"),
    "p_float_pinf": float("Infinity"),
    "p_float_minf": float("-Infinity"),
    "p_blob": b"abc123",
    "p_uuid": UUID("9c5b94b1-35ad-49bb-b118-8e8fc24abf80"),
    "p_decimal": decimal.Decimal("123.456"),
    "p_date": DataAPIDate.from_string("+11111-09-30"),
    "p_duration": DataAPIDuration(signum=+1, months=1, days=1, nanoseconds=60000000001),
    "p_inet": ipaddress.ip_address("10.1.1.2"),
    "p_time": DataAPITime(12, 34, 56, 789120000),
    "p_timestamp": DataAPITimestamp.from_string("2015-05-03T13:30:54.234Z"),
    "p_timestamp_out_offset": DataAPITimestamp.from_string(
        "-0123-04-03T13:13:04.123+1:00"
    ),
    "p_list_int": [99, 100, 101],
    "p_list_double": [
        1.1,
        2.2,
        float("NaN"),
        float("Infinity"),
        float("-Infinity"),
        9.9,
    ],
    "p_set_ascii": DataAPISet(["a", "b", "c"]),
    "p_set_float": DataAPISet([1.1, float("NaN"), float("-Infinity"), 9.9]),
    "p_map_text_float": DataAPIMap({"a": 0.1, "b": 0.2}),
    "p_map_float_text": DataAPIMap({1.1: "1-1", float("NaN"): "NANNN!"}),
    "somevector": DataAPIVector.from_bytes(
        convert_ejson_binary_object_to_bytes({"$binary": "PczMzb5MzM0+mZma"})
    ),
    "embeddings": DataAPIVector([0.1, -0.2, 0.3]),
    "p_counter": 100,
    "p_timeuuid": UUID("0de779c0-92e3-11ef-96a4-a745ae2c0a0b"),
}

OUTPUT_NONCUSTOMTYPES_ROW_TO_POSTPROCESS = {
    "p_timestamp": "2015-05-03T13:30:54.234Z",
    "somevector": {"$binary": "PczMzb5MzM0+mZma"},  # [0.1, -0.2, 0.3] but not bit-wise
    "embeddings": [0.1, -0.2, 0.3],
    "p_date": "1956-09-30",
    "p_duration": "P1DT2M3S",
    "p_time": "12:34:56.78912",
    "p_set_ascii": ["a", "b", "c"],
    "p_set_float": [1.1, "NaN", "-Infinity", 9.9],
    "p_map_text_float": {"a": 0.1, "b": 0.2},
    "p_map_float_text": {1.1: "1-1", "NaN": "NANNN!"},
}

EXPECTED_NONCUSTOMTYPES_POSTPROCESSED_ROW = {
    "p_timestamp": datetime.datetime(
        2015, 5, 3, 13, 30, 54, 234000, tzinfo=datetime.timezone.utc
    ),
    "somevector": bytes_to_floats(
        convert_ejson_binary_object_to_bytes({"$binary": "PczMzb5MzM0+mZma"})
    ),
    "embeddings": [0.1, -0.2, 0.3],
    "p_date": datetime.date(1956, 9, 30),
    "p_duration": datetime.timedelta(days=1, minutes=2, seconds=3),
    "p_time": datetime.time(12, 34, 56, 789120),
    "p_set_ascii": {"a", "b", "c"},
    "p_set_float": {1.1, float("NaN"), float("-Infinity"), 9.9},
    "p_map_text_float": {"a": 0.1, "b": 0.2},
    "p_map_float_text": {1.1: "1-1", float("NaN"): "NANNN!"},
}

OUTPUT_FILLERS_ROW_TO_POSTPROCESS: dict[str, Any] = {}
EXPECTED_FILLERS_POSTPROCESSED_ROW: dict[str, Any] = {
    "p_text": None,
    "p_boolean": None,
    "p_int": None,
    "p_float": None,
    "p_float_nan": None,
    "p_float_pinf": None,
    "p_float_minf": None,
    "p_blob": None,
    "p_uuid": None,
    "p_decimal": None,
    "p_date": None,
    "p_duration": None,
    "p_inet": None,
    "p_time": None,
    "p_timestamp": None,
    "p_timestamp_out_offset": None,
    "p_list_int": [],
    "p_list_double": [],
    "p_set_ascii": DataAPISet(),
    "p_set_float": DataAPISet(),
    "p_map_text_float": DataAPIMap({}),
    "p_map_float_text": DataAPIMap({}),
    "somevector": None,
    "embeddings": None,
    "p_counter": None,
    "p_timeuuid": None,
}
EXPECTED_FILLERS_NONCUSTOMTYPES_POSTPROCESSED_ROW: dict[str, Any] = {
    "p_text": None,
    "p_boolean": None,
    "p_int": None,
    "p_float": None,
    "p_float_nan": None,
    "p_float_pinf": None,
    "p_float_minf": None,
    "p_blob": None,
    "p_uuid": None,
    "p_decimal": None,
    "p_date": None,
    "p_duration": None,
    "p_inet": None,
    "p_time": None,
    "p_timestamp": None,
    "p_timestamp_out_offset": None,
    "p_list_int": [],
    "p_list_double": [],
    "p_set_ascii": set(),
    "p_set_float": set(),
    "p_map_text_float": {},
    "p_map_float_text": {},
    "somevector": None,
    "embeddings": None,
    "p_counter": None,
    "p_timeuuid": None,
}


INPUT_ROW_TO_PREPROCESS = {
    "none": None,
    "str": "the str",
    "int": 987,
    "bool": True,
    "float": 123.4,
    "float_nan": float("NaN"),
    "float_pinf": float("Infinity"),
    "float_minf": float("-Infinity"),
    "bytes": b"\xcd\xcc\xcc={\x00\x00\x00",  # {'$binary': 'zczMPXsAAAA='}
    "DataAPITimestamp": DataAPITimestamp(-65403045296110),  # "-103-06-17T07:25:03.890Z"
    "DataAPIDate": DataAPIDate(1724, 4, 22),  # Kant's date of birth
    "DataAPITime": DataAPITime(4, 5, 6, 789000000),
    "d.datetime": datetime.datetime(
        1724, 4, 22, 4, 5, 6, 789000, tzinfo=datetime.timezone.utc
    ),
    "d.date": datetime.date(1724, 4, 22),
    "d.time": datetime.time(4, 5, 6, 789000),
    "Decimal": decimal.Decimal("123.456"),
    "Decimal_nan": decimal.Decimal("NaN"),
    "Decimal_pinf": decimal.Decimal("Infinity"),
    "Decimal_ninf": decimal.Decimal("-Infinity"),
    #
    "DataAPIVector": DataAPIVector([0.1, -0.2, 0.3]),
    "DataAPIMap_f": DataAPIMap(
        [
            (float("NaN"), 1.0),
            ("k2", float("NaN")),
            ("k3", "v3"),
        ],
    ),
    "dict_f": {
        float("NaN"): 1.0,
        "k2": float("NaN"),
        "k3": "v3",
    },
    "list_f": [0.1, float("Infinity"), 0.3],
    "DataAPISet_f": DataAPISet([0.1, float("Infinity"), 0.3]),
    "set_f": {float("Infinity")},  # 1-item, because of regular sets being unordered
}

EXPECTED_PREPROCESSED_ROW = {
    "none": None,
    "str": "the str",
    "int": 987,
    "bool": True,
    "float": 123.4,
    "float_nan": "NaN",
    "float_pinf": "Infinity",
    "float_minf": "-Infinity",
    "bytes": {"$binary": "zczMPXsAAAA="},
    "DataAPITimestamp": "-0103-06-17T07:25:03.890Z",
    "DataAPIDate": "1724-04-22",
    "DataAPITime": "04:05:06.789",
    "d.datetime": "1724-04-22T04:05:06.789000+00:00",
    "d.date": "1724-04-22",
    "d.time": "04:05:06.789000",
    "Decimal": decimal.Decimal("123.456"),
    "Decimal_nan": "NaN",
    "Decimal_pinf": "Infinity",
    "Decimal_ninf": "-Infinity",
    #
    "DataAPIVector": {"$binary": "PczMzb5MzM0+mZma"},  # [0.1, -0.2, 0.3],
    "DataAPIMap_f": {
        "NaN": 1.0,
        "k2": "NaN",
        "k3": "v3",
    },
    "dict_f": {
        "NaN": 1.0,
        "k2": "NaN",
        "k3": "v3",
    },
    "list_f": [0.1, "Infinity", 0.3],
    "DataAPISet_f": [0.1, "Infinity", 0.3],
    "set_f": ["Infinity"],
}

ROGUE_TABLE_DESCRIPTION = {
    "name": "table_rogue",
    "definition": {
        "columns": {
            "col_static_list": {
                "type": "list",
                "apiSupport": {
                    "createTable": False,
                    "insert": True,
                    "read": True,
                    "filter": False,
                    "cqlDefinition": "static list<int>",
                },
            },
            "col_static_list_exotic": {
                "type": "list",
                "apiSupport": {
                    "createTable": False,
                    "insert": True,
                    "read": True,
                    "filter": False,
                    "cqlDefinition": "static list<blob>",
                },
            },
            "col_static_map": {
                "type": "map",
                "apiSupport": {
                    "createTable": False,
                    "insert": True,
                    "read": True,
                    "filter": False,
                    "cqlDefinition": "static map<int, text>",
                },
            },
            "col_static_map_exotic": {
                "type": "map",
                "apiSupport": {
                    "createTable": False,
                    "insert": True,
                    "read": True,
                    "filter": False,
                    "cqlDefinition": "static map<blob, blob>",
                },
            },
            "col_static_set": {
                "type": "set",
                "apiSupport": {
                    "createTable": False,
                    "insert": True,
                    "read": True,
                    "filter": False,
                    "cqlDefinition": "static set<int>",
                },
            },
            "col_static_set_exotic": {
                "type": "set",
                "apiSupport": {
                    "createTable": False,
                    "insert": True,
                    "read": True,
                    "filter": False,
                    "cqlDefinition": "static set<blob>",
                },
            },
            "col_static_timestamp": {
                "type": "timestamp",
                "apiSupport": {
                    "createTable": False,
                    "insert": True,
                    "read": True,
                    "filter": True,
                    "cqlDefinition": "static timestamp",
                },
            },
            "col_unsupported": {
                "type": "UNSUPPORTED",
                "apiSupport": {
                    "createTable": False,
                    "insert": False,
                    "read": False,
                    "filter": False,
                    "cqlDefinition": (
                        "frozen<list<frozen<map<frozen<set<float>>, smallint>>>>"
                    ),
                },
            },
        },
        "primaryKey": {"partitionBy": [], "partitionSort": {}},
    },
}

ROGUE_OUTPUT_ROW_TO_POSTPROCESS = {
    "col_static_list": [1, 2, 3],
    "col_static_list_exotic": [
        {"$binary": "/w=="},
        {"$binary": "/w=="},
    ],
    "col_static_map": [
        [1, "one"],
    ],
    "col_static_map_exotic": [
        [{"$binary": "/w=="}, {"$binary": "/w=="}],
    ],
    "col_static_set": [1, 2, 3],
    "col_static_set_exotic": [
        {"$binary": "/w=="},
    ],
    "col_static_timestamp": "2022-01-01T12:34:56Z",
    "col_unsupported": {"not", "really:", "just", "testing", "passthrough", 123},
}

# Only for very specific cases of these columns do we expect a proper parsing:
ROGUE_EXPECTED_POSTPROCESSED_ROW = {
    "col_static_list": [1, 2, 3],
    "col_static_list_exotic": [
        {"$binary": "/w=="},
        {"$binary": "/w=="},
    ],
    "col_static_map": [
        [1, "one"],
    ],
    "col_static_map_exotic": [
        [{"$binary": "/w=="}, {"$binary": "/w=="}],
    ],
    "col_static_set": [1, 2, 3],
    "col_static_set_exotic": [
        {"$binary": "/w=="},
    ],
    "col_static_timestamp": DataAPITimestamp.from_string(
        ROGUE_OUTPUT_ROW_TO_POSTPROCESS["col_static_timestamp"]  # type: ignore[arg-type]
    ),
    "col_unsupported": {"not", "really:", "just", "testing", "passthrough", 123},
}


class TestTableConverters:
    @pytest.mark.describe("test of row postprocessors from schema")
    def test_row_postprocessors_from_schema(self) -> None:
        col_desc = ListTableDescriptor.coerce(TABLE_DESCRIPTION)
        tpostprocessor = create_row_tpostprocessor(
            columns=col_desc.definition.columns,
            options=defaultSerdesOptions.with_override(
                SerdesOptions(
                    custom_datatypes_in_reading=True,
                    datetime_tzinfo=None,
                ),
            ),
            similarity_pseudocolumn=None,
        )

        converted_column = tpostprocessor(OUTPUT_ROW_TO_POSTPROCESS)
        assert _repaint_NaNs(converted_column) == _repaint_NaNs(
            EXPECTED_POSTPROCESSED_ROW
        )

        with pytest.raises(ValueError):
            tpostprocessor({"bippy": 123})

    @pytest.mark.describe("test of row postprocessors from schema, no custom types")
    def test_row_postprocessors_from_schema_nocustom(self) -> None:
        col_desc = ListTableDescriptor.coerce(TABLE_DESCRIPTION)
        tpostprocessor = create_row_tpostprocessor(
            columns={
                c_n: c_d
                for c_n, c_d in col_desc.definition.columns.items()
                if c_n in EXPECTED_NONCUSTOMTYPES_POSTPROCESSED_ROW
            },
            options=defaultSerdesOptions.with_override(
                SerdesOptions(
                    custom_datatypes_in_reading=False,
                    datetime_tzinfo=datetime.timezone.utc,
                ),
            ),
            similarity_pseudocolumn=None,
        )

        converted_column = tpostprocessor(OUTPUT_NONCUSTOMTYPES_ROW_TO_POSTPROCESS)
        assert _repaint_NaNs(converted_column) == _repaint_NaNs(
            EXPECTED_NONCUSTOMTYPES_POSTPROCESSED_ROW
        )

    @pytest.mark.describe(
        "test of primary-key postprocessors based on pk-schema, custom types"
    )
    def test_pk_postprocessors_from_schema_custom(self) -> None:
        # constructing a fake "primary key list" and associated test assets
        # from the full 'row' for the row-postprocess test.
        # Careful in the different input format for blobs
        primary_key_schema = {
            col_name: col_dict
            for col_name, col_dict in ListTableDescriptor.coerce(
                TABLE_DESCRIPTION
            ).definition.columns.items()
        }
        primary_key_list = [
            (
                OUTPUT_ROW_TO_POSTPROCESS[col_name]["$binary"]  # type: ignore[index]
                if isinstance(col_desc, TableScalarColumnTypeDescriptor)
                and col_desc.column_type == ColumnType.BLOB
                else OUTPUT_ROW_TO_POSTPROCESS[col_name]
            )
            for col_name, col_desc in primary_key_schema.items()
        ]
        expected_primary_key = EXPECTED_POSTPROCESSED_ROW

        ktpostprocessor = create_key_ktpostprocessor(
            primary_key_schema=primary_key_schema,
            options=defaultSerdesOptions.with_override(
                SerdesOptions(
                    custom_datatypes_in_reading=True,
                    datetime_tzinfo=None,
                ),
            ),
        )

        assert _repaint_NaNs(ktpostprocessor(primary_key_list)[1]) == _repaint_NaNs(
            expected_primary_key
        )

        with pytest.raises(ValueError):
            ktpostprocessor(["bippy", 123])

    @pytest.mark.describe(
        "test of primary-key postprocessors based on pk-schema, no custom types"
    )
    def test_pk_postprocessors_from_schema_nocustom(self) -> None:
        primary_key_schema = {
            col_name: col_dict
            for col_name, col_dict in ListTableDescriptor.coerce(
                TABLE_DESCRIPTION
            ).definition.columns.items()
            if col_name in EXPECTED_NONCUSTOMTYPES_POSTPROCESSED_ROW
        }
        primary_key_list = [
            (
                OUTPUT_NONCUSTOMTYPES_ROW_TO_POSTPROCESS[col_name]["$binary"]  # type: ignore[index]
                if isinstance(col_desc, TableScalarColumnTypeDescriptor)
                and col_desc.column_type == ColumnType.BLOB
                else OUTPUT_NONCUSTOMTYPES_ROW_TO_POSTPROCESS[col_name]
            )
            for col_name, col_desc in primary_key_schema.items()
        ]
        expected_primary_key = EXPECTED_NONCUSTOMTYPES_POSTPROCESSED_ROW

        ktpostprocessor = create_key_ktpostprocessor(
            primary_key_schema=primary_key_schema,
            options=defaultSerdesOptions.with_override(
                SerdesOptions(
                    custom_datatypes_in_reading=False,
                    datetime_tzinfo=datetime.timezone.utc,
                ),
            ),
        )

        assert _repaint_NaNs(ktpostprocessor(primary_key_list)[1]) == _repaint_NaNs(
            expected_primary_key
        )

    @pytest.mark.describe("test of type-based row preprocessor")
    def test_row_preprocessors_from_types(self) -> None:
        ptp_opts = defaultSerdesOptions.with_override(
            SerdesOptions(
                binary_encode_vectors=True,
                unroll_iterables_to_lists=True,
            ),
        )
        preprocessed_row = preprocess_table_payload(
            INPUT_ROW_TO_PREPROCESS,
            options=ptp_opts,
            map2tuple_checker=None,
        )
        assert preprocessed_row == EXPECTED_PREPROCESSED_ROW

        # unroll-iterables option
        gen_row_0 = {"gen_col": (i for i in range(5))}
        preprocessed_gen_0 = preprocess_table_payload(
            gen_row_0,
            options=ptp_opts,
            map2tuple_checker=None,
        )
        assert preprocessed_gen_0 == {"gen_col": [0, 1, 2, 3, 4]}
        gen_row_1 = {"gen_col": (i for i in range(5))}
        preprocessed_gen_1 = preprocess_table_payload(
            gen_row_1,
            options=defaultSerdesOptions.with_override(
                SerdesOptions(unroll_iterables_to_lists=False),
            ),
            map2tuple_checker=None,
        )
        assert preprocessed_gen_1 == gen_row_1

        # vector-encoding
        vec_data = [0.1, -0.2, 0.3]
        dvec_row = {"dvec": DataAPIVector(vec_data)}
        preprocessed_dvec_0 = preprocess_table_payload(
            dvec_row,
            options=ptp_opts,
            map2tuple_checker=None,
        )
        assert preprocessed_dvec_0 == {"dvec": {"$binary": "PczMzb5MzM0+mZma"}}
        preprocessed_dvec_1 = preprocess_table_payload(
            dvec_row,
            options=defaultSerdesOptions.with_override(
                SerdesOptions(binary_encode_vectors=False),
            ),
            map2tuple_checker=None,
        )
        assert preprocessed_dvec_1 == {"dvec": vec_data}

        with pytest.raises(ValueError):
            preprocess_table_payload(
                {"err_field": ObjectId()},
                options=ptp_opts,
                map2tuple_checker=None,
            )

    @pytest.mark.describe("test of row postprocessors from schema, fillers")
    def test_row_postprocessors_from_schema_fillers(self) -> None:
        col_desc = ListTableDescriptor.coerce(TABLE_DESCRIPTION)
        tpostprocessor = create_row_tpostprocessor(
            columns=col_desc.definition.columns,
            options=defaultSerdesOptions.with_override(
                SerdesOptions(custom_datatypes_in_reading=True),
            ),
            similarity_pseudocolumn=None,
        )

        converted_column = tpostprocessor(OUTPUT_FILLERS_ROW_TO_POSTPROCESS)
        assert _repaint_NaNs(converted_column) == _repaint_NaNs(
            EXPECTED_FILLERS_POSTPROCESSED_ROW
        )

    @pytest.mark.describe(
        "test of row postprocessors from schema, fillers, no custom types"
    )
    def test_row_postprocessors_from_schema_fillers_nocustom(self) -> None:
        col_desc = ListTableDescriptor.coerce(TABLE_DESCRIPTION)
        tpostprocessor = create_row_tpostprocessor(
            columns=col_desc.definition.columns,
            options=defaultSerdesOptions.with_override(
                SerdesOptions(custom_datatypes_in_reading=False),
            ),
            similarity_pseudocolumn=None,
        )

        converted_column = tpostprocessor(OUTPUT_FILLERS_ROW_TO_POSTPROCESS)
        assert _repaint_NaNs(converted_column) == _repaint_NaNs(
            EXPECTED_FILLERS_NONCUSTOMTYPES_POSTPROCESSED_ROW
        )

    @pytest.mark.describe("test of row postprocessors for rogue table")
    def test_row_postprocessors_rogue_table(self) -> None:
        """Rogue table: columns that are unsupported and/or not fully parseable."""
        col_desc = ListTableDescriptor.coerce(ROGUE_TABLE_DESCRIPTION)
        tpostprocessor = create_row_tpostprocessor(
            columns=col_desc.definition.columns,
            options=defaultSerdesOptions,
            similarity_pseudocolumn=None,
        )

        converted_column = tpostprocessor(ROGUE_OUTPUT_ROW_TO_POSTPROCESS)
        assert _repaint_NaNs(converted_column) == _repaint_NaNs(
            ROGUE_EXPECTED_POSTPROCESSED_ROW
        )

        with pytest.raises(ValueError):
            tpostprocessor({"bippy": 123})
